#!/usr/bin/env python

"""Simple command line tool for viewing trees in browser.

Commands
>>> toytree draw [options] STDIN              # return as HTML
>>> toytree root [options] STDIN              # root and return to nwk
>>> toytree distance [options] STDIN1 STDIN2  # compare trees
>>> toytree io [options] STDIN                # read/write to nwk

Chain commands
>>> toytree root -o A B C $file | toytree draw - -o /tmp/tree.html
>>> toytree draw -o /tmp/test.html -ts p '((a,b),c)'
>>> toytree root \
>>>   https://eaton-lab.org/data/Cyathophora.tre -o "prz*" -r | \
>>>   toytree draw -v -ts o -
"""

from typing import Optional
import argparse
import sys
import tempfile
import subprocess
from pathlib import Path
import toytree

TMPFILE = Path(tempfile.gettempdir()) / "test"
DESCRIPTION = "toytree command line tool. Select a subcommand."
EPILOG = "EXAMPLE:\n$ toytree draw TREE -ts o -d 400 400 -v"


def string_or_stdin_parse(intree) -> str:
    """If TREE is stdin then return the string from stdin."""
    if intree == "-":
        return sys.stdin.read().strip()
    return intree


def setup_parsers() -> argparse.ArgumentParser:
    """Setup and return an ArgumentParser w/ subcommands."""
    parser = argparse.ArgumentParser(
        prog="toytree",
        description=DESCRIPTION,
        epilog=EPILOG,
    )
    parser.add_argument("-v", "--version", action='version', version=f"toytree {toytree.__version__}")
    subparsers = parser.add_subparsers(help="sub-commands", dest="subcommand")

    # toytree draw [options] TREE --------------------------------------------
    parser_draw = subparsers.add_parser(
        "draw", 
        help="create tree drawing",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_draw.add_argument(
        "-ts", type=str, default="n", metavar="treestyle", help="tree style",
    )
    parser_draw.add_argument(
        "-d", type=int, metavar="dim", nargs=2, default=(None, None),
        help="width height (px)",
    )
    parser_draw.add_argument(
        "-o", type=Path, metavar="basename", default=TMPFILE,
        help="output basename[.format suffix]",
    )
    parser_draw.add_argument(
        "-v", type=str, metavar="app", nargs="?", const="auto",
        help="open file with default browser or app.",
    )
    parser_draw.add_argument(
        "-f", type=str, help="output file format", default="html",
        choices=["html", "svg", "pdf"]
    )
    parser_draw.add_argument(
        "TREE", type=string_or_stdin_parse, help="tree newick file or string",
    )

    # toytree root [options] TREE --------------------------------------------
    parser_root = subparsers.add_parser("root", help="(re)root tree and return to STDOUT")
    parser_root.add_argument("-o", type=str, nargs="+", default=(), help="outgroup")
    # parser_root.add_argument("-r", action="store_true", help="use regex matching on outgroup string.")
    parser_root.add_argument("TREE", type=str, help="tree newick file or string")

    # toytree distance [options] TREE TREE ----------------------------------
    parser_distance = subparsers.add_parser("distance", help="compute distance between trees")
    parser_distance.add_argument("TREE1", type=str, help="tree1 newick file or string")
    parser_distance.add_argument("TREE2", type=str, help="tree2 newick file or string")
    parser_distance.add_argument(
        "-m", type=str,
        choices=['rf', 'rfi', 'rfj', 'qrt'],
        default='rf',
        help="distance metric method",
    )
    parser_distance.add_argument(
        "-n", "--normalize", action="store_true",
        help="normalize value between [0-1]")
    return parser


def main(cmd: Optional[str] = None) -> int:
    """Command line tool."""
    parser = setup_parsers()
    args = parser.parse_args(cmd.split() if cmd else None)

    # root
    if args.subcommand == "root":
        tree = toytree.tree(args.TREE)
        tree = tree.root(*args.o)
        sys.stdout.write(tree.write(None))
        return 0

    # draw
    if args.subcommand == "draw":
        # print(args)
        tree = toytree.tree(args.TREE)
        width, height = args.d
        canvas, _, _ = tree.draw(tree_style=args.ts, width=width, height=height)
        canvas.style['background-color'] = "white"

        path = Path(args.o).with_suffix(f".{args.f}")
        if args.f == "html":
            import toyplot.html
            toyplot.html.render(canvas, str(path))
        if args.f == "svg":
            import toyplot.svg
            toyplot.svg.render(canvas, str(path))
        if args.f == "pdf":
            import toyplot.pdf
            toyplot.pdf.render(canvas, str(path))
        if args.v:
            if args.v == "auto":
                import webbrowser
                webbrowser.open(str(path))
            else:
                subprocess.run([str(args.v), str(path)], check=True)
        return 0

    # distance
    if args.subcommand == "distance":
        tree1 = toytree.tree(args.TREE1)
        tree2 = toytree.tree(args.TREE2)
        norm = {"normalize": args.normalize}
        if args.m == "rf":
            val = toytree.distance.get_treedist_rf(tree1, tree2, **norm)
        elif args.m == "rfg_mci":
            val = toytree.distance.get_treedist_rfg_mci(tree1, tree2, **norm)
        elif args.m == "rfg_spi":
            val = toytree.distance.get_treedist_rfg_spi(tree1, tree2, **norm)
        elif args.m == "rfi":
            val = toytree.distance.get_treedist_rfi(tree1, tree2, **norm)
        elif args.m == "qrt":
            val = toytree.distance.get_treedist_qrt(tree1, tree2)
        sys.stdout.write(val)
        return 0

    # unreachable
    parser.print_help()
    return 1


def test():
    parser = setup_parsers()
    # print(parser.parse_args(["--version"]))
    # print(parser.parse_args(["--help"]))
    # print(parser.parse_args("root ((a,b),c); -o a".split()))
    # print(parser.parse_args("root ((a,b),c);".split()))
    print(parser.parse_args("root -o a b ((a,b),c);".split()))
    # parser.parse_args(["root", "--help"])


if __name__ == "__main__":

    main()



#!/usr/bin/env python

"""Core ToyTree class object of the toytree package.

Examples
--------
>>> tree = toytree.rtree.unitree(10)
>>> tree = toytree.ToyTree(toytree.Node('a'))
>>> tree = toytree.tree("((a,b),c);")

References
----------
- toytree
"""

from __future__ import annotations
from typing import (
    Sequence, Dict, List, Optional, Iterator, Any, Union, Tuple,
    TypeVar, Set,  # Callable,
)
import re
from pathlib import Path
from copy import deepcopy
from hashlib import md5
# from collections.abc import Sequence as SequenceType

from loguru import logger
import numpy as np
from toyplot import Canvas
from toyplot.coordinates import Cartesian

# subpackage object APIs
from toytree.core.apis import (
    TreeModAPI, TreeDistanceAPI, TreeEnumAPI, PhyloCompAPI, AnnotationAPI)
from toytree.core.node import Node
from toytree.style import TreeStyle
from toytree.drawing import draw_toytree, ToyTreeMark
from toytree.utils.src.exceptions import (
    ToytreeError, NODE_NOT_IN_TREE_ERROR, NODE_INDEXING_ERROR)
import toytree
# from toytree.io.src.writer import write_newick

# toytree logger
logger = logger.bind(name="toytree")

# Type alias for Node selection
Query = TypeVar("Query", str, int, Node)
Color = TypeVar("Color", str, np.ndarray, tuple)  # toyplot.ColorMap, ...
UNPACKING_MSG = """\
Use unpacking on collections:
>>> query_list = [0, 1, 2]
>>> tree.method(*query_list, ...)"""


class ToyTree:
    """ToyTree class for manipulating and drawing trees.

    ToyTrees should generally be created using a constructor function
    such as `toytree.tree` or `toytree.rtree`, to init a tree from
    input data (e.g. newick) or random generators, respectively. This
    class can be used for type checking.

    Parameters
    ----------
    treenode: Node
        A toytree.Node class instance representing the tree root.

    Attributes
    ----------
    nnodes: int
        Number of Nodes in the tree.
    ntips: int
        Number of leaf Nodes (tips) in the tree. Updated on modification.
    features:
        The features names assigned to Nodes. A dynamic property.
    edge_features
        Set of feature names applying to edges, not nodes.
    style: TreeStyle
        Mutable dataclass with base drawing style.
    mod: TreeModAPI
        API to apply methods from `toytree.mod` to this tree.
    enum: TreeEnumAPI
        API to apply methods from `toytree.enum` to this tree.
    distance: TreeDistanceAPI
        API to apply methods from `toytree.distance` to this tree.
    pcm: PhyloCompAPI
        API to apply methods from `toytree.pcm` to this tree.
    annotate: AnnotationAPI
        API to apply methods from `toytree.annotate` to this tree.

    Examples
    --------
    >>> tree = toytree.tree("((a,b),c);")
    >>> isinstance(tree, toytree.ToyTree)
    >>> # True
    """
    def __init__(self, treenode: Node) -> ToyTree:
        """Initialize a ToyTree from a Node instance."""

        self.treenode = treenode
        self.nnodes: int = 0
        self.ntips: int = 0
        self.style = TreeStyle()
        self.edge_features: Set = set(("dist", "support"))
        self._idx_dict: Dict[int, Node] = {}
        """Private dict mapping Node idx labels to Node instances."""

        # toytree subpackage library API (mod, pcm, distance, layout)"""
        self.mod = TreeModAPI(self)
        self.distance = TreeDistanceAPI(self)
        self.pcm = PhyloCompAPI(self)
        self.enum = TreeEnumAPI(self)
        self.annotate = AnnotationAPI(self)

        # update Node idxs, _idx_dict, nnodes, ntips, and Node heights
        self._update()

    #####################################################
    # DUNDERS
    #####################################################
    # def __len__(self) -> int:
        # """Ambiguous, does one expect ntips or nnodes? So it is
        # not supported. See .ntips and .nnodes attrs."""
        # return self.ntips

    def __iter__(self) -> Iterator[Node]:
        """ToyTree is iterable, returning Nodes in idx order."""
        return (self[i] for i in range(self.nnodes))

    # def __getitem__(self, idx: int) -> Node:
    #     """Nodes can be accessed by indexing or slicing by idx label"""
    #     # allow indexing by int, e.g., [3]
    #     try:
    #         return self._idx_dict[idx]
    #     except Exception:
    #         pass
    #     # allow slicing by ints, e.g., [3:10:2]
    #     try:
    #         return [self._idx_dict[idx] for idx in range(*idx.indices(self.nnodes))]
    #     except Exception:
    #         pass
    #     # allow indexing by a Sequence, e.g., [3, 10, 2, 4]
    #     try:
    #         return [self._idx_dict[i] for i in idx]
    #     except Exception:
    #         pass
    #     # if a negative number then get positive and reindex
    #     try:
    #         if isinstance(idx, int) and idx < 0:
    #             return self._idx_dict[self.nnodes + idx]  # idx is negative
    #     # raise a helpful error message.
    #     except Exception as exc:
    #         raise ToytreeError(NODE_INDEXING_ERROR) from exc

    def __getitem__(self, idx: int) -> Node:
        """Nodes can be accessed by indexing or slicing by idx label"""
        # allow indexing by int, e.g., [3]
        try:
            # try casting to int, to support int, np.int64, np.int32, etc
            # if isinstance(idx, int):
            try:
                idx = int(idx)
                if idx >= 0:
                    return self._idx_dict[idx]
                else:
                    return self._idx_dict[self.nnodes + idx]  # idx is negative
            except (ValueError, TypeError):
                pass

            # allow indexing by a Sequence, e.g., [3, 10, 2, 4]
            if isinstance(idx, (list, np.ndarray)):
                return [self._idx_dict[i] for i in idx]

            # allow slicing by ints, e.g., [3:10:2]
            else:
                if hasattr(idx, "indices"):
                    return [self._idx_dict[idx] for idx in range(*idx.indices(self.nnodes))]
                raise ToytreeError("invalid indexing type.")

        # raise a helpful error message.
        except Exception as exc:
            raise ToytreeError(NODE_INDEXING_ERROR) from exc

    def __repr__(self) -> str:
        """Short object representation for toytree.core.tree.ToyTree"""
        return f"<toytree.ToyTree at {hex(id(self))}>"

    # def __str__(self) -> str:
    #     """Return ascii representation of tree."""
    #     return "\n".join(self.treenode._get_ascii()[0])

    @property
    def nedges(self) -> int:
        """Return the number of edges in the tree, *not including the
        root edge if the tree is rooted*.

                        |  <- (not counted)
                       _T_
                     _|_  |                  ___T_
                    |   | |                 |  |  |
                    A   B C                 A  B  C
              (rooted w/ 4 edges)      (unrooted with 3 edges)

        Note
        ----
        This value is not cached and requires an iteration of the tree
        in order to accommodate counting edges of unresolved nodes.
        For super speed-sensitive operations use nnodes-1 if tree is
        bifurcating, else measure the value once and cache it if the
        tree doesn't change.

        See Also
        --------
        `ToyTree.iter_edges`
            Generator of edges as child-parent Node pairs. Has the
            option `include_root` to toggle counting of root split.

        Example
        -------
        >>> tree = toytree.rtree.unittree(5, seed=123)
        >>> tree.nedges == sum(1 for i in tree.iter_edges())
        >>> # True
        """
        internal = sum(1 for i in self.enum.iter_edges())
        return internal
        # to add counting of the root edge of rooted trees do this.
        # return internal + (1 if self.is_rooted() else 0)

    #####################################################
    # FEATURES
    #####################################################

    @property
    def features(self) -> Tuple[str]:
        """Return a tuple of all Node data feature names.

        The basic features present on all Nodes include 'dist',
        'name', 'support', 'height', and 'idx'. These features will
        all be shown if you call `ToyTree.get_node_data()`. To add
        additional features to use `ToyTree.set_node_data`, or add as
        attributes to Node objects. If features apply to edges instead
        of Nodes (e.g., dist, support) they are also listed in
        ToyTree.edge_features.

        Notes:
        This function finds node features dynamically by visiting every
        Node in the tree. It is thus not performant for speed sensitive
        code.

        Examples
        --------
        >>> tree = toytree.rtree.unittree(10)
        >>> # set 'color' feature on a single Node.
        >>> tree[5].color = 'red'
        >>> # get 'color' values for all Nodes.
        >>> tree.get_node_data("color", missing="blue")
        """
        feats = set()
        for node in self:
            feats.update(node.__dict__)
        feats = (i for i in feats if not i.startswith("_"))
        defaults = ("idx", "name", "height", "dist", "support")
        return defaults + tuple(sorted(feats))

    def remove_features(self, *feature: str, inplace: bool = False) -> ToyTree:
        """Remove one or more non-default data features from all Nodes.

        This function is very rarely needed. Note that you cannot
        remove default features "idx", "name", "height", "dist", or
        "support".

        Examples
        --------
        >>> tree = tree.set_node_data("new_feature", {0: 10})
        >>> tree = tree.remove_features("new_feature")
        """
        tree = self if inplace else self.copy()
        for feat in feature:
            if feat in ("idx", "name", "height", "dist", "support"):
                raise toytree.utils.NodeDataError(
                    f"Cannot remove required Node feature: {feature}")
            for node in tree.traverse():
                delattr(node, feat)
        return tree

    #####################################################
    # IDENTITY
    #####################################################

    def is_rooted(self) -> bool:
        """Return False if the tree is unrooted."""
        if len(self.treenode.children) > 2:
            return False
        return True

    def is_bifurcating(self, include_root: bool = True) -> bool:
        """Return False if no polytomies exist in tree.

        Parameters
        ----------
        include_root: bool
            If False then the state of the root node is ignored when
            checking for polytomies.
        """
        tris = [len(node.children) <= 2 for node in self]
        if include_root:
            return all(tris)
        return all(tris[:-1])

    def copy(self) -> ToyTree:
        """Return a deepcopy of the ToyTree."""
        return deepcopy(self)
        # return ToyTree(self.treenode.copy())

    #####################################################
    # TRAVERSAL
    # Visit all connected Nodes, and/or create ._idx_dict cache.
    #####################################################

    def traverse(self, strategy: str = "levelorder") -> Iterator[Node]:
        """Return an iterator over Nodes in a specific traversal order.

        Traversal strategies: admonition
        idxorder:
            Leaf nodes are visited first from left to right, followed
            by internal nodes in postorder traversal. Note: this
            traversal order is already cached on any ToyTree and can
            thus be accessed much faster by indexing.
        preorder:
            Parents are visited before children. Traverses all the way
            down each left subtree before proceeding to right child.
        postorder:
            Children are visited before parents. The left subtree is
            visited, then right, then the parent.
        levelorder:
            Nodes the same distance (number of edges) from root are
            visited left to right, before descending to next level.
        inorder:
            Nodes are visited in non-descreasing order if they are
            a binary search tree: left child, parent, right child.

        Parameters
        ----------
        strategy: str
            A traversal strategy for the order in which nodes will
            be visited: 'preorder', 'postorder', 'levelorder',
            'inorder', or 'idxorder'.

        Examples
        --------
        Generate tree drawings enumerating each traversal strategy
        >>> tree = toytree.rtree.unittree(10, seed=123)
        >>> ords = ['postorder', 'preorder', 'levelorder', 'idxorder']
        >>> for trav in ords:
        >>>     for idx, node in enumerate(tree.traverse(trav)):
        >>>         node.traverse_order = str(idx)
        >>>     _, a, _ = tree.draw(
        >>>         node_labels="traverse_order",
        >>>         node_sizes=16, node_mask=False);
        >>>     a.label.text = trav
        """
        for node in self.treenode.traverse(strategy=strategy):
            yield node

    def _update(self) -> None:
        """Traverse to set and cache Node idxorder and coordinates.

        idxorder traversal is used to fill the ToyTree._idx_dict to
        make Nodes easily indexable. While doing this it also calculates
        Node heights and spacing (_x) and counts nnodes and ntips.

        If a topology has been modified then idx labels must be
        updated. This function is called by all internal `toytree.mod`
        functions which modify the topology (e.g., add_child, root,
        rotate, etc) but not if users modify Nodes adhoc. This is why
        Node objects are immutable.
        """
        # clear depth counters used to get heights during traversal
        depths = {self.treenode: 0.}

        # queue starts with root children, and stack starts with root.
        queue = list(self.treenode._children)

        # start w/ root on stack, as either an inner or a tip.
        if queue:
            inner_stack = [self.treenode]
            outer_stack = []
        else:
            inner_stack = []
            outer_stack = [self.treenode]

        # traverse left then right subtrees to fill and pull from queue
        while queue:
            # get node from start of queue to proceed levelorder
            node = queue.pop()

            # set depth of this node from the root
            depths[node] = depths[node._up] + node._dist

            # if leaf add to output stack and update farthest depth
            if node.is_leaf():
                outer_stack.append(node)
            else:
                inner_stack.append(node)

            # add node's children to the queue (left child on end)
            queue.extend(node._children)

        # get max_depth from root, height is measured relative to this.
        max_depth = max(depths.values())

        # clear idx cache and counter to be filled next
        idx = 0
        self._idx_dict.clear()

        # return nodes in reverse order they were added to stack
        while outer_stack:
            node = outer_stack.pop()
            node._height = max_depth - depths[node]
            node._x = idx
            node._idx = idx
            self._idx_dict[idx] = node
            idx += 1
        self.ntips = idx

        # return internal nodes, or just root if only a single Node.
        while inner_stack:
            node = inner_stack.pop()
            node._height = max_depth - depths[node]
            node._x = sum(i._x for i in node._children) / len(node._children)
            node._idx = idx
            self._idx_dict[idx] = node
            idx += 1
        self.nnodes = idx

    #####################################################
    # TREE MODIFICATION FUNCTIONS (See ToyTree.mod)
    # - root, unroot, rotate_node, ladderize,
    # - set_node_heights, collapse_nodes, prune,
    # - drop_tips, resolve_polytomy,
    # - root, unroot
    #####################################################

    #################################################
    # NODES SEARCH/MATCH BY FLEXIBLE INPUTS
    # Matching Nodes by name can be used to color nodes/edges...
    #################################################

    def _iter_nodes_by_name_match(self, *query: str) -> Iterator[Node]:
        """Return Iterator over Nodes in idxorder matched by leaf names
        while allowing for regular expression matched of names.
        """
        # compile regex strings
        comps = []
        for que in query:
            if not que.startswith("~"):
                comps.append((que, None))
            else:
                try:
                    comps.append((que, re.compile(que[1:])))
                except re.error as exc:
                    msg = f"invalid regex query {query} raised re.error:\n{exc}"
                    logger.error(msg)
                    raise ToytreeError(msg) from exc

        # traverse tree in idxorder checking each Node for a name match
        matched = set()
        for node in self:
            for que, regex in comps:
                if regex:
                    # match() is faster but search() is more flexible...
                    # if regex.match(node.name):
                    if regex.search(node.name):
                        matched.add(que)
                        yield node
                else:
                    if que == node.name:
                        yield node
                        matched.add(que)

        # raise exception for non-matched queries
        not_matched = set(query) - set(matched)
        if not_matched:
            msg = f"No Node names match query: {not_matched}"
            logger.error(msg)
            raise ValueError(msg)

    def get_nodes(self, *query: Query) -> List[Node]:
        """Return a list of Nodes matching a flexible Query.

        One or more Node instances are returned that match one ore more
        Node Queries. A Node Query can be an int, str, or Node type,
        and is matched depending on the type. Nodes are simply returned,
        int matches to Node idx labels, and str matches to Node name
        features. A str prefixed by `~` can match multiple Node names
        as a regular expression search. (e.g., `~r[0-3]`) matches will
        match ["r0", "r1", "r2", "r3"]. If no query is entered then
        all Nodes are returned.

        You can [practice your regex here](https://pythex.org/).
        The order in which Nodes are returned may be different than the
        order of queries.

        This function is used inside many other `toytree` functions that
        accept `Query` as an argument.

        Parameters
        ----------
        query: str, int, Node, or None
            Flexible query selector can search for Nodes by name, idx
            label, or by entering a Node directly. Multiple values can
            be entered to return a list of matching Nodes.

        Examples
        --------
        >>> tree = toytree.rtree.unittree(16, seed=123)
        >>> tree.get_nodes()                    # all Nodes returned
        >>> tree.get_nodes("r1")                # [Node(1)]
        >>> tree.get_nodes("r1", "r2")          # [Node(1), Node(2)]
        >>> tree.get_nodes("~r[1-2]$")          # [Node(1), Node(2)]
        >>> tree.get_nodes(5, 6)                # [Node(5), Node(6)]
        >>> tree.get_nodes(tree[5], "r1")       # [Node(5), Node(1)]
        >>> tree.get_nodes(*[2,3])              # [Node(2), Node(3)]
        """
        nodes = set()
        names = set()
        for que in query:
            if isinstance(que, int):
                nodes.add(self[que])
            elif isinstance(que, Node):
                if que not in self:
                    raise ValueError(NODE_NOT_IN_TREE_ERROR)
                nodes.add(que)
            elif isinstance(que, str):
                names.add(que)  # names are expanded to Nodes below.
            elif isinstance(que, np.integer):
                nodes.add(self[que])
            else:
                msg = f"query type {type(que)} not supported. "
                if isinstance(que, (list, tuple)):
                    msg += UNPACKING_MSG
                logger.error(msg)
                raise TypeError(msg)

        # match Node names as a group so we only need to perform one
        # tree traversal. Each query can return multiple regex hits.
        matched = set(self._iter_nodes_by_name_match(*names))
        nodes.update(matched)

        # if not query then return all Nodes
        if not nodes:
            return list(self)
        return list(nodes)

    def get_mrca_node(self, *query: Query) -> Node:
        """Return the MRCA Node from Nodes matching a flexible query.

        Find and return the most-recent-common-ancestor Node instance
        based on input selectors that can be either Node names, Node
        int idx labels, or Node objects. Input types are detected
        automatically and handled, so that even mixed input types can
        be entered (e.g., `get_mrca_node('a', 2, 3)`). If the selected
        Nodes do not share a common ancestor this will raise a
        ToytreeError. This function is useful for selecting and
        annotating clades on a tree drawing. Node names can also be
        entered as regular expressions to match multiple names by
        prefixing the name with "~" (e.g., .get_mrca_node("~r[0-3]$"))

        Parameters
        ----------
        *query: str, int, Node
            One or more Node objects, Node str names, or Node int idx
            labels, any of which can be used to select Nodes.

        Notes
        -----
        If the tree is unrooted then the mrca will be found relative
        to the 'pseudo-root' (the Node that exists as parent to the
        basal trichotomy, but has no parent associated with it).

        Examples
        --------
        >>> tree = toytree.rtree.unittree(12)
        >>> mrca = tree.get_mrca_node('r1', 'r2', 'r3')
        >>> mrca = tree.get_mrca_node('r[1-3]$')
        >>> mrca = tree.get_mrca_node(12)
        >>> mrca = tree.get_mrca_node(12, 13, 14, 15)
        >>> mrca = tree.get_mrca_node(*(tree[i] for i in (12, 13)))
        >>> print(mrca.idx, mrca.get_descendants())
        """
        # get flexible input as Node instances
        nodes = self.get_nodes(*query)
        if len(nodes) == 1:
            return nodes[0]
        # include_self necessary to find ancestor between tip ^ parent
        nset = set.intersection(*(
            set(i.iter_ancestors(include_self=True)) for i in nodes)
        )
        return min(nset)

    def get_ancestors(
        self,
        *query: Query,
        include_query: bool = True,
        include_top: bool = True,
        stop_at_mrca: bool = False,
    ) -> List[Node]:
        """Return a set of Nodes that are ancestors of the query samples.

        The returned set can include or exclude the sample query; it
        can trace back all ancestors to the root of the tree, or only
        to the MRCA of the sample query; and it can include or exclude
        the top node (root or MRCA depending on arguments).

        Parameters
        ----------
        *query: str, int, Node
            One or more Node objects, Node str names, or Node int idx
            labels, any of which can be used to select Nodes.
        include_query: bool
            If False the query Nodes are not included in returned set.
        include_top: bool
            If False the 'top' Node of the set is not included. This
            can be either the tree root, or the MRCA Node, depending
            on the `stop_at_mrca` argument.
        stop_at_mrca: bool
            If False then all ancestors are included back to the root
            of the tree. If True, ancestors only trace back to MRCA
            of the sample query.

        Note
        ----
        See also `Node.get_ancestors` which will fetch the ancestors of
        an individual Node. By contrast, this function returns the set
        of Nodes ancestral to a group of one or more queried Nodes,
        and has additional start,stop criteria arguments. To get a
        single Node's ancestors this method is slightly slower, but is
        faster for finding shared ancestors.

        Examples
        --------
        >>> Draw a tree and color ancestors of Nodes 1, 2, 3
        >>> tree = toytree.rtree.unittree(ntips=8, seed=123)
        >>> ancs = tree.get_ancestors(1,2,3)
        >>> tree.set_node_data("path", {i: 1 for i in ancs}, inplace=True)
        >>> tree.draw(ts='p', node_colors=("path", "Set2"));
        """
        # expand query into a set of Nodes
        query = set(self.get_nodes(*query))

        # get union of the ancestors of each query Node, not including the Node
        ancestors = set.union(*[set(i.get_ancestors()) for i in query])

        # optionally include query nodes in ancestor set.
        if include_query:
            ancestors.update(query)

        # stopping criterion (global root or sample mrca)
        if stop_at_mrca:
            mrca = self.get_mrca_node(*query)
            ancestors = {i for i in ancestors if i._idx <= mrca._idx}
            if not include_top:
                ancestors.discard(mrca)
        else:
            if not include_top:
                ancestors.discard(self.treenode)
        return ancestors

    def get_node_mask(
        self,
        *unmask: Query,
        show_tips: bool = False,
        show_internal: bool = False,
        show_root: bool = False,
    ) -> np.ndarray:
        """Return a boolean array masking all Nodes except selected.

        Creates a boolean mask to hide a set of selected Nodes.
        The array is in Node idxorder (from 0-nnodes) where boolean
        True will *mask* Nodes, and False will *show* Nodes. Additional
        Nodes can be selected to be *unmasked* by entering Node int idx
        labels or name strings. The default mask is an array that
        masks tip Nodes but unmasks all internal Nodes.

        Parameters
        ----------
        *unmask: int, str or Node
            Select Nodes using a Query (int or str labels, including
            regular expressions) that will be unmasked (shown).
        show_tips: bool
            If True all tip Nodes will be masked.
        show_internal: bool
            If True all internal Nodes will be masked.
        show_root: bool
            If True the root Node will be masked.

        Examples
        --------
        >>> tree = toytree.rtree.unittree(10, seed=123)
        >>> # create a mask that only shows tips & Nodes 15 and 16
        >>> mask = tree.get_node_mask(15, 16, show_tips=True)
        >>> tree.draw(ts='s', node_mask=mask);
        """
        # default is all False (mask all)
        arr = np.zeros(self.nnodes, dtype=np.bool_)
        if show_tips:
            arr[:self.ntips] = True
        if show_internal:
            arr[self.ntips:-1] = True
        if show_root:
            arr[-1] = True
        # check unmask this way because 0 matches a Node.
        if unmask != ():
            # supports regex expansion
            for node in self.get_nodes(*unmask):
                arr[node.idx] = True
        return arr

    def is_monophyletic(self, *query: Query) -> bool:
        """Return True if leaf Nodes form a monophyletic clade.

        If any other leaf Nodes are members of this clade, but not
        included in the input set of 'nodes', then these Nodes are
        not monophyletic. Nodes can be entered either as Node objects
        or by their int idx labels.

        Parameters
        ----------
        *query: Node, str, or int
            One or more Node objects, Node name str, or Node idx int
            labels to check for monophyly. If None then all Nodes are
            used as the query (returns True).

        Examples
        --------
        >>> tree = toytree.rtree.unittree(ntips=10, seed=123)
        >>> tree.draw()
        >>> print(tree.is_monophyletic(0, 1, 2))
        >>> print(tree.is_monophyletic(0, 4, 8))
        """
        if not self.is_rooted():
            raise ToytreeError("The tree must be rooted to test monophyly")
        nodes = self.get_nodes(*query)
        mrca = self.get_mrca_node(*nodes)
        for node in mrca.iter_leaves():
            if node not in nodes:
                return False
        return True

    ##################################################
    # TREE DISTANCE (ToyTree.distance)
    # get distances between nodes, trees, or data points
    ##################################################

    ###################################################
    # I/O FORMATTING (toytree.io)
    # read and write trees to newick, nexus, nhx
    ###################################################

    ###################################################
    # TOPOLOGY OR LEAF ANALYSIS FUNCTIONS
    # access nodes or features ...
    ###################################################
    def iter_tip_labels(self) -> Iterator[str]:
        """Generator of tip labels in idx order."""
        for node in self[:self.ntips]:
            yield node._name

    def get_tip_labels(self) -> List[str]:
        """Return a list of tip labels in Node idx order."""
        return list(self.iter_tip_labels())

    ###################################################
    # TOPOLOGY ENUMERATION
    # access info about splits, bipartitions, quartets, etc.
    # from toytree.enum subpackage:
    # - iter_edges
    # - iter_bipartitions
    # - iter_quartets
    ###################################################

    def get_topology_id(self, feature="name", include_root: bool = False) -> str:
        """Return a unique ID representing this topology.

        Two trees with the same topology and tip names will produce
        the same id, i.e., the rotation of Nodes does not affect the
        generated ID. Rooting/Unrooting does affect it. The ID string
        is useful for identifying unique topologies among a set of
        trees without requiring distance comparisons. This method
        uses an md5 hash of a string of ordered Node names with
        the digest value represented as a string of hexadecimal digits.

        Parameters
        ----------
        features: str
            The feature used to represent tip Nodes (default='name').
            This should be a feature that is unique among tip Nodes,
            and is relevant to identifying similarity among the trees
            you plan to compare using topology id strings. Careful
            changing this option from 'name' unless you are familiar
            with the consequences.
        include_root: bool
            By default the root Node is excluded (if tree is rooted)
            such that all unrooted trees with the same toplogy return
            the same topology_id. To distinguish among differently
            rooted versions of the same tree set `include_root=True`.

        Examples
        --------
        >>> tree.get_topology_id() # '70f5cfb041f176d86020971ac5f633e1'

        See Also
        --------
        - iter_bipartitions
        """
        # bipartitions are ordered by edge idx order, and names within
        # bipartitions are consistently ordered by alphanumeric names.
        # so we sort so that trees with the same topology but rotated
        # nodes will have the same bipartitions.
        biparts = sorted(self.iter_bipartitions(
            feature=feature, sort=True, type=tuple))

        # optional: duplicate last bipart to indicate rooting
        if include_root and self.is_rooted():
            biparts += [
                tuple(sorted(
                    sorted(i.get_leaf_names()) for i in self.treenode.children
                ))
            ]
        return md5(str(biparts).encode('utf-8')).hexdigest()

    ###################################################
    # COORDINATE LAYOUT FUNCTIONS
    # push to .layout subpackage
    ###################################################

    def _iter_node_coordinates(self) -> Iterator[Tuple[float, float]]:
        """Generator of 'unstyled' cached node coordinates."""
        for node in self:
            yield node._x, node._height

    # def get_node_coordinates(self, **kwargs) -> pd.DataFrame:
    #     """Return a DataFrame with xy coordinates for plotting nodes.

    #     This returns coordinates that could be used when adding
    #     additional annotations to plots, such as scatterplot points,
    #     or error bars on top of nodes. By default Node idx=0 will be
    #     located at coordinate position (0, 0), which can be modified
    #     using the `xbaseline` and `ybaseline` args.

    #     Take care when calling this function that the coordinates
    #     will be different depending on the *style* arguments applied.
    #     The style args come from the `.style` dict-like object of the
    #     ToyTree, and can be overriden by additional args to this func,
    #     the same as in the `.draw()` function. For example, layout
    #     facing down ('d') will yield different coordinates than layout
    #     facing up ('u').

    #     Examples
    #     --------
    #     >>> style = {'layout': 'd', 'xbaseline': 10}
    #     >>> canvas, axes, mark = tree.draw(**style)
    #     >>> node_coords = tree.get_node_coordinates(**style)
    #     >>> axes.scatterplot(coords.x, coords.y, size=10);
    #     """
    #     style = get_tree_style_base(self, **kwargs)
    #     style = validate_style(self, style, **kwargs)
    #     coords = get_layout(self, style).coords
    #     data = pd.DataFrame(
    #         columns=('x', 'y'),
    #         index=range(self.nnodes),
    #         data=coords,
    #     )
    #     return data

    # def get_tip_coordinates(self, **kwargs) -> pd.DataFrame:
    #     """Return a DataFrame with xy coordinates for tip nodes.

    #     See `ToyTree.get_node_coordinates` for details.
    #     """
    #     return self.get_node_coordinates(**kwargs).iloc[:self.ntips]

    ###################################################
    # FULL TREE DATA GET/SET
    # functions to modify features of all connected Nodes
    # see source in toytree/data/
    # - set_node_data
    # - set_node_data_from_dataframe
    # - get_node_data
    # - get_tip_data
    ###################################################

    def get_feature_dict(self, keys: str = None, values: str = None) -> Dict[str, Any]:
        """Return a dict mapping selected Node features as keys, values.

        This can be used to return a dict mapping any two arbitrary
        features, or to Node objects. Examples include mapping Node
        objects to dist values, or idx labels to Node names, or
        Node names to Node objects. There are many possibilities. The
        value of None for `keys` or `values` returns Node objects.
        Retured dict is in idxorder.

        Parameters
        ----------
        keys: str or None
            Select the keys of the returned dictionary.
        values: str or None
            Select the values of the returned dictionary.

        Examples
        --------
        >>> nodes_to_dists = tree.get_feature_dict(None, 'dist')
        >>> idx_to_names = tree.get_feature_dict('idx', 'name')
        >>> names_to_nodes = tree.get_feature_dict('name', None)

        See Also
        --------
        get_node_data
            Return feature data as a DataFrame w/ options for
            how to impute values for missing features on some nodes.
        """
        ndict = {}
        try:
            for _, node in self._idx_dict.items():
                if keys is not None:
                    key = getattr(node, keys)
                else:
                    key = node
                if values is not None:
                    value = getattr(node, values)
                else:
                    value = node
                ndict[key] = value
        except AttributeError as exc:
            raise ToytreeError(
                f"feature_dict cannot build {keys} -> {values} mapping "
                "because one or\nmore of the selected features is not assigned "
                "to every Node.\nSee `get_node_data()` for working with "
                "missing values."
            ) from exc

        # check that keys were not duplicated
        if len(ndict) != self.nnodes:
            raise ToytreeError(
                f"feature_dict cannot be built because {keys} "
                "does not have unique values, and thus Nodes with the "
                "same value cannot be represented as keys in the dict.")
        return ndict

    ###################################################
    # DRAWING
    ###################################################

    # ------------------------------------------------------------------
    # Draw functions imported, but docstring here...
    # ------------------------------------------------------------------
    def _draw_browser(self, *args, new: bool = False, tmpdir: Path = None, **kwargs):
        """Open and display tree drawing in default web browser.

        TODO: overload toyplot function, option to reuse same tab,
        add div styling, etc.
        Or, maybe make this at toytree level as `toytree.draw(canvas)`
        also make a `toytree.save()` shortcut to saving in formats.
        """
        # import toyplot.browser
        canvas, axes, mark = self.draw(**kwargs)
        toytree.utils.show([canvas], new=new, tmpdir=tmpdir)
        return canvas, axes, mark

    def draw(
        self,
        tree_style: Optional[str] = None,
        height: int = None,
        width: int = None,
        axes: Cartesian = None,
        layout: str = None,
        tip_labels: Union[bool, Sequence] = None,
        tip_labels_colors: Union[Color, Sequence[Color]] = None,
        tip_labels_angles: Union[float, Sequence[float]] = None,
        tip_labels_style: Dict[str, Any] = None,
        tip_labels_align: bool = None,
        node_mask: Union[bool, Sequence[bool]] = None,
        node_labels: Union[bool, Sequence[str]] = None,
        node_labels_style: Dict[str, Any] = None,
        node_sizes: Union[int, Sequence[int]] = None,
        node_colors: Union[str, Sequence[str]] = None,
        node_style: Dict[str, Any] = None,
        node_hover: bool = None,
        node_markers: Sequence[str] = None,
        node_as_edge_data: bool = False,
        edge_colors: Union[str, Sequence[str]] = None,
        edge_widths: float = None,
        edge_type: str = None,
        edge_style: Dict[str, Any] = None,
        edge_align_style: Dict[str, Any] = None,
        edge_markers: Sequence[str] = None,
        edge_labels: Union[bool, Sequence[str]] = None,
        use_edge_lengths: bool = None,
        scale_bar: bool = None,
        padding: float = None,
        xbaseline: float = None,
        ybaseline: float = None,
        admixture_edges: List[Tuple[int, int]] = None,
        shrink: float = None,
        fixed_order: Sequence[str] = None,
        fixed_position: Sequence[float] = None,
        label: Optional[str] = None,
        **kwargs,
    ) -> Tuple[Canvas, Cartesian, ToyTreeMark]:
        """Return a drawing of the tree as a Toyplot figure.

        Drawings are returned as Tuple[Canvas, Cartesian, ToytreeMark]
        following the style of the toyplot plotting library and will
        render as html figures automatically in jupyter notebook.
        The Canvas can be saved to file in a number of formats using
        toyplot. The Cartesian axes object can be used to position
        multiple Marks onto the same cartesian coordinates, and to
        style axes ticks and labels (see docs). The Mark object can
        be further modified to edit or access style args.

        Parameters
        ----------
        tree_style: str or None
            Select a builtin base TreeStyle on top of which to add other
            style modifications. Options include "n", "c", "p", "o",
            "r", "m", "d"; you can alo create your own (see docs).
            Using a tree_style overrides any `ToyTree.style` dict.
        ts: str or None
            A shorter alias that can be used for `tree_style`.
        height: int or None
            If None the plot height is auto-sized. If 'axes' arg is
            used tree is drawn on an existing Axes and this arg is
            ignored. Else, this sets height of the Canvas in px units.
        width: int
            If None the plot width is auto-sized. If 'axes' arg is
            used tree is drawn on an existing Axes and this arg is
            ignored. Else, this sets width of the Canvas in px units.
        axes: toyplot.coordinates.Cartesian
            A toyplot Cartesian axes object. If provided, tree is drawn
            on it. If not provided then a new Canvas and Cartesian
            axes are created and returned with the tree Mark added to
            it. See documentation for details on how this option is
            used to create composite drawings combining tree plots
            with other data plots (e.g., scatterplots, barplots).
        layout: str or None
            Layout defines the direction parent-child relationships are
            drawn. Options are 'r' 'l', 'u', 'd' for 'right', 'left',
            'up', or 'down' respectively. 'c' draws circular trees from
            angle 0-360 degrees, or cX-Y draws circular trees on an arc
            from X degrees to Y degrees. Finally, any other entry, such
            as None, draws an 'unrooted' layout. Default='r'.
        tip_labels: bool or Sequence[str]
            If True, tip labels ('name' features on tip Nodes) are
            added to the plot; if False, no tip labels are added. If a
            list of tip labels is provided it must be the same len as
            ntips and is applied in Node idx order (0-ntips).
        tip_labels_colors: Color or Sequence[Color] or (feature, colormap)
            A valid toyplot Color, Sequence[Color] of len=ntips, or
            tuple of (feature, colormap) to map colors to tip labels.
        tip_labels_style: Dict[str, str]
            A dictionary of CSS style arguments to apply to text
            tip labels. See tree.style for options.
        tip_labels_angles: int, Sequence[int], or None
            None will set angles based on layout method.
        tip_labels_align: bool
            If True tip names will be aligned and dashed edges will
            drawn to extend from tree edges to the tip names.
        node_mask: bool or Sequence[bool]
            Masks nodes (size, color, shape, label) if True, shows
            nodes if False. An iterable can be entered to selectively
            hide some nodes. The convenience function .get_node_mask()
            can be used to generate mask arrays. Default options
            vary among tree styles, but usually hide tip nodes.
        node_labels: bool, str, or Sequence[str]
            Labels associated with nodes. True shows node idx labels,
            False hides node labels (sets to ""). A string or
            iterable of strings assigns labels to nodes 0-nnodes.
            An iterable of string values can be generated from node
            features using .get_node_data() or .get_node_labels(),
            the latter includes string formatting options.
        node_sizes: int or Sequence[int]
            Size of node markers can be set as an integer or Iterable
            of integers in node order 0-nnodes. Node size 0 is hidden.
            The node_mask argument sets nodes to size 0 when masked,
            and overrides this argument.
        node_colors: str or Sequence[str]
            Color of node markers can be a single color or Iterable
            of colors in node order 0-nnodes. Any valid toyplot color
            (str, rgb array, rgba array, hex, etc) is accepted. See
            the toyplot.color module. The default color palette is
            accessible from toytree.colors. If all nodes will be set
            to the same color it is more efficient to use the
            node_style dictionary (node_style={"fill": 'red'}). If
            used, node_colors overrides 'fill' in node_style.
        node_style: Dict[str, str]
            A dict of valid CSS styles to apply to node markers, such
            as 'fill', 'stroke'. See tree.style for options.
        node_hover: True, False, or Sequence[str]
            Default is True in which case node hover will show the
            node values. If False then no hover is shown. If a list or
            dict is provided (which should be in node order) then the
            values will be shown in order. If a dict then labels can
            be provided as well.
        node_markers: str or Sequence[str]
            The shape of node markers: 'o'=circle, 's'=square. See
            toyplot documentation for all available options:
            https://toyplot.readthedocs.io/en/stable/markers.html
        node_as_edge_data: bool
            If True then node_markers and node_labels are instead
            plotted as edge_markers and edge_labels. For more fine
            control see `toytree.annotate`.
        edge_colors: str or Sequence[str]
            A color or collection of colors nnodes in length to apply
            to edges in node idx order.
        edge_widths: float or Sequence[float]
            A width arg in px units, or collection of widths to apply
            to edges in node idx order.
        edge_type: str
            Edges can be phylogram ('p') or cladogram ('c') type.
        edge_style: Dict[str, Any]
            A dictionary of valid CSS style args to apply to edge
            lines. See tree.style for available options.
        edge_align_style: Dict[str, Any]
            A dictionary of valid CSS style args to apply to aligned
            edge lines. See tree.style for available options.
        use_edge_lengths: bool
            If True edge lengths ('dist' features of TreeNodes) are
            represented in drawings. If False all terminal edges are
            extended to align tips at 0.
        scale_bar: bool
            If True then the axis corresponding to the height of the
            tree will be set to visible and tick marks will be auto-
            generated to span from time=0 to root height. The style
            of the axes can be further modified from the axes object
            after the draw function is called.
        padding: float
            Padding space between the drawing and the visible axes. This
            is a setting of the Cartesian axes and can be modified more
            after plotting. Default=20 (px).
        xbaseline: float
            Shift the position of the tree along x-axis.
        ybaseline: float
            Shift the position of the tree along y-axis.
        admixture_edges: Tuple, List[Tuple]
            Admixture edges add colored edges to the plot in the
            style of the 'edge_align_style'. These will be drawn
            from (source, dest, height, width, color). Example:
            `>>> [(4, 3, 50000, 3, 'red')]`.
        fixed_order: Sequence[str]
            An Iterable of tip labels in the order they should be
            plotted. The default is the node names in idx order
            0-ntips. These nodes will be plotted on the coordinates
            0-ntips on either the x or y-axis depending on the
            layout of the tree drawing. This is a convenient argument
            for visualizing discordance of trees.
        fixed_position: Sequence[float]
            The positions on the tip axis where ordered tips should
            be plotted. If None then default positions range(0, ntips)
            are used. By setting explicit positions tips can be
            spaced to better show discordance, or extinct taxa, or
            morphology. The order of positions applies to tips in
            order of the 'fixed_order' arg.

        Examples
        --------
        >>> tree = toytree.rtree.unittree(ntips=10)
        >>> tree.draw();
        >>> canvas, axes, mark = tree.draw(ts="o", scale_bar=True);
        >>>
        >>> # save drawing to file.
        >>> import toyplot.svg
        >>> toyplot.svg.render(canvas, "saved-plot.svg")
        """
        kwargs = dict(
            # toytree=self,
            tree_style=tree_style,
            axes=axes,
            height=height,
            width=width,
            layout=layout,
            tip_labels=tip_labels,
            tip_labels_colors=tip_labels_colors,
            tip_labels_align=tip_labels_align,
            tip_labels_angles=tip_labels_angles,
            tip_labels_style=tip_labels_style,
            node_mask=node_mask,
            node_labels=node_labels,
            node_labels_style=node_labels_style,
            node_sizes=node_sizes,
            node_hover=node_hover,
            node_style=node_style,
            node_colors=node_colors,
            node_markers=node_markers,
            node_as_edge_data=node_as_edge_data,
            edge_type=edge_type,
            edge_colors=edge_colors,
            edge_widths=edge_widths,
            edge_style=edge_style,
            edge_align_style=edge_align_style,
            # edge_labels=edge_labels,                 # test
            # edge_markers=edge_markers,               # test
            # edge_marker_colors=edge_marker_colors,   # test
            # edge_marker_sizes=edge_marker_size,      # test
            # edge_marker_mask=edge_marker_mask,       # test
            use_edge_lengths=use_edge_lengths,
            scale_bar=scale_bar,
            padding=padding,
            xbaseline=xbaseline,
            ybaseline=ybaseline,
            admixture_edges=admixture_edges,
            shrink=shrink,
            fixed_order=fixed_order,
            fixed_position=fixed_position,
            label=label,
            kwargs=kwargs,
        )
        # private debugging mode returns just the kwargs
        if kwargs.get("kwargs").get("debug"):
            return kwargs

        # draw the ToyTree
        try:
            canvas, axes, mark = draw_toytree(tree=self, **kwargs)
            return canvas, axes, mark
        except ToytreeError as exc:
            logger.error(exc)
            raise exc
        except Exception as exc:
            logger.error(exc)
            raise exc


# def inside_notebook() -> bool:
#     """Return True if executed from inside jupyter, else False.

#     takes only ~140 ns.
#     """
#     try:
#         shell = get_ipython().__class__.__name__
#         if shell == 'ZMQInteractiveShell':
#             return True   # Jupyter notebook or qtconsole
#         elif shell == 'TerminalInteractiveShell':
#             return False  # Terminal running IPython
#         else:
#             return False  # Other type (?)
#     except NameError:
#         return False      # Probably standard Python interpreter


if __name__ == "__main__":

    # import toytree
    tree_ = toytree.rtree.unittree(12, treeheight=1232344, seed=123)
    # tree = tree_.mod.edges_slider(0.5)
    # c, a, m = tree_._draw_browser(tree_style='s', layout='d', new=False)
    # print(tree_.write(dist_formatter="%.12g"))
    # print(tree_.get_node_data())
    # print(tree_.treenode)
    # tree_.draw()
    print(tree_[80])

    # print(tree.get_tip_labels())

    # tree = tree.set_node_data("color", {i: "red" for i in (2,3,4)})
    # print(tree[3].color)
    # print(tree.features)
    # print(tree.get_node_data())
    # print(tree.get_tip_data("height"))
    # print(tree.get_node_mask())
    # print(tree.get_bipartitions(exclude_singleton_splits=False))
    # print(tree._get_bipartitions_table(exclude_singleton_splits=True))



#!/usr/bin/env python

import re


def replace_whitespace(nwk: str, sub: str = "") -> str:
    """Replace whitespace with sub unless preceded by a semicolon.

    This gets rid of whitespace within newick strings while retaining
    newlines separating newick strings for parsing multitrees.
    """
    pattern = r'(?<!;)\s'
    modified_nwk = re.sub(pattern, sub, nwk)
    return modified_nwk.strip()



#!/usr/bin/env python

"""...

Priority in color/opacity
-------------------------
1. color settings can take a single or multiple colors. If multiple
colors are entered then individual colors are assigned to each item,
but if only one then the color is set to None and a value will be set
to fill/stroke instead.
>>> color = [a, b, c, ...]

2. 
>>> fill/stroke = x


3. Color rgba value is the default opacity value used. But if user sets
teh 

"""

from __future__ import annotations
from typing import Union, Sequence, TypeVar, Tuple, Optional, Iterator, Any, List
from dataclasses import dataclass, field
from copy import deepcopy
from loguru import logger
from toytree.utils import ToytreeError

logger = logger.bind(name="toytree")
Color = TypeVar("Color")
ToyTree = TypeVar("ToyTree")


@dataclass(repr=False, init=True, eq=False)
class SubStyle:
    """A subclass of Style for CSS on markers."""

    def __delattr__(self, key) -> None:
        raise ToytreeError("TreeStyle dict keys cannot be deleted.")

    def __repr__(self):
        """Return a serialized JSON formatted style dict."""
        block = ["{"]
        for key, val in self.__dict__.items():
            block.append(f"    {key}: {val!r},")
        block.append("}")
        return "\n".join(block)

    def __getitem__(self, key):
        """Get item using Python name, fetch from dict as CSS name."""
        if "-" in key:
            if key == "-toyplot-anchor-shift":
                key = "anchor_shift"
            else:
                key = key.replace("-", "_")
        return self.__dict__[key]

    def __setitem__(self, key, value):
        """Set item using Python name, set to dict as CSS name."""
        if "-" in key:
            if key == "-toyplot-anchor-shift":
                key = "anchor_shift"
            else:
                key = key.replace("-", "_")
        self.__dict__[key] = value

    def copy(self) -> SubStyle:
        """Return a deepcopy."""
        return deepcopy(self)


@dataclass(repr=False, init=True, eq=False)
class NodeStyle(SubStyle):
    fill: Color = (0.4, 0.7607843137254902, 0.6470588235294118, 1.0)
    fill_opacity: float = None
    stroke: Color = "#262626"
    stroke_width: float = 1.5
    stroke_opacity: float = None


@dataclass(repr=False, init=True, eq=False)
class NodeLabelStyle(SubStyle):
    fill: Color = (0.145, 0.145, 0.145, 1.0)
    fill_opacity: float = 1.0
    font_size: Union[int, str] = 9
    font_weight: int = 300
    font_family: str = "Helvetica"
    anchor_shift: Union[str, int] = 0
    baseline_shift: Union[str, int] = 0
    text_anchor: str = "middle"


@dataclass(repr=False, init=True, eq=False)
class EdgeStyle(SubStyle):
    stroke: Color = (0.145, 0.145, 0.145, 1.0)
    stroke_width: float = 2.0
    stroke_opacity: Optional[float] = None
    stroke_linecap: str = "round"
    stroke_dasharray: Optional[str] = None


@dataclass(repr=False, init=True, eq=False)
class EdgeAlignStyle(SubStyle):
    stroke: Color = (0.66, 0.66, 0.66, 1)
    stroke_width: int = 2
    stroke_opacity: Optional[float] = 0.75
    stroke_linecap: str = "round"
    stroke_dasharray: str = "2,4"


@dataclass(repr=False, init=True, eq=False)
class TipLabelStyle(SubStyle):
    fill: Color = (0.145, 0.145, 0.145, 1.0)
    fill_opacity: Optional[float] = None
    font_size: Union[str, float] = 12
    font_weight: int = 300
    font_family: str = "Helvetica"
    anchor_shift: Union[str, float] = 15
    baseline_shift: Union[str, int] = 0
    text_anchor: str = "start"


@dataclass(repr=False, init=True, eq=False)
class TreeStyle:
    """The base tree style on top of which other tree_styles are defined."""
    tree_style: str = None
    """: A tree_style name to set a base style."""
    height: int = None
    """: Canvas height in px units."""
    width: int = None
    """: Canvas width in px units."""
    layout: str = "r"
    """: Tree layout method name for projecting tree in coordinate space."""
    edge_type: str = "p"
    """: Edge type defines straight, diagonal, or curved edges."""
    edge_colors: Union[Color, Sequence[Color], Tuple[str]] = None
    """: Edges can be all the same or different colors."""
    edge_widths: Union[float, Sequence[float], Tuple[str]] = None
    """: Edges can be all the same or different widths in px units."""

    node_mask: Union[bool, Sequence[bool], Tuple[int, int, int]] = None
    """: Node labels are str plotted on top of Node markers."""
    node_colors: Union[Color, Sequence[Color], Tuple[str]] = None
    """: Node colors can set different 'fill' to Node markers."""
    node_sizes: Union[float, Sequence[float], Tuple[str]] = 0.0
    """: Node sizes can set different size to Node markers."""
    node_markers: Union[str, Sequence[str]] = "o"
    """: Node markers define the shape of Node markers."""
    node_hover: Union[bool, str, Sequence[str]] = None
    """: Node hover creates a tooltip for interactive data inspection."""
    node_labels: Union[bool, str, Sequence[str], Tuple[str]] = False
    """: Node labels are str plotted on top of Node markers."""
    node_as_edge_data: bool = False
    """: Node markers and labels are shown as edge markers and labels."""

    tip_labels: Union[bool, Sequence[str], Tuple[str]] = True
    """: Tip labels are str plotted below leaf Node markers."""
    tip_labels_colors: Union[Color, Sequence[Color], Tuple[str]] = None
    """: Tip labels colors can be same or different for each tip."""
    tip_labels_angles: Union[float, Sequence[float]] = None
    """: Tip labels angles can be set but are usually auto-generated."""
    tip_labels_align: bool = None
    """: Align tip labels at farthest Node from root."""

    edge_style: EdgeStyle = field(default_factory=EdgeStyle)
    """: SubStyle dict for setting edge width, stroke, etc."""
    node_style: NodeStyle = field(default_factory=NodeStyle)
    """: SubStyle dict for setting Node fill, stroke, opacity, etc."""
    node_labels_style: NodeLabelStyle = field(default_factory=NodeLabelStyle)
    """: SubStyle dict for setting Node label font, size, fill, etc."""
    tip_labels_style: TipLabelStyle = field(default_factory=TipLabelStyle)
    """: SubStyle dict for setting Tip label font, size, fill, etc."""
    edge_align_style: EdgeAlignStyle = field(default_factory=EdgeAlignStyle)
    """: SubStyle dict for setting Aligned tip labels edges stroke, etc."""

    # self.show_root_edge: bool = None
    use_edge_lengths: bool = True
    """: Plot topology with transformed edge lengths for easier viewing."""
    scale_bar: Union[bool, float] = False
    """: Modify Cartesian axes to show scale in 'dist' feature units."""
    padding: float = 15.0
    """: Modify Cartesian axes to set padding between Mark and axes in px units."""
    xbaseline: float = 0.0
    """: Shift tree on Cartesian axes so Node 0 is at (xbaseline, ybaseline)."""
    ybaseline: float = 0.0
    """: Shift tree on Cartesian axes so Node 0 is at (xbaseline, ybaseline)."""
    shrink: float = 0.0
    """: Add additional space to fit tip names relative to tree."""
    admixture_edges: List[Tuple] = None

    def __repr__(self):
        """Return a serialized JSON formatted style dict."""
        block = ["{"]
        for key, val in self.__dict__.items():
            if isinstance(val, SubStyle):
                block.append(f"{key}: {val!r},")
            else:
                block.append(f"{key}: {val!r},")
        block.append("}")
        return "\n".join(block)

    def __iter__(self) -> Iterator[str]:
        """ToyTree is iterable, returning Nodes in idx order."""
        return (i for i in self.__dict__)

    def _items(self) -> Iterator[Tuple[str, Any]]:
        return ((i, j) for i, j in self.__dict__.items())

    def copy(self) -> TreeStyle:
        """Return a deepcopy."""
        return deepcopy(self)


if __name__ == "__main__":

    ts = TreeStyle()

    ts.tip_labels_style.font_size = 15
    ts.tip_labels_style["font_size"] = 16
    ts.tip_labels_style["font-size"] = 17

    print(ts.tip_labels_style)
    print(ts.tip_labels_style.font_size)



#!/usr/bin/env python

"""Possible Network Class Implementation. 
    This file also contains old versions of network/src/parse_network functions
"""

from typing import Union, Tuple, List
from pathlib import Path
import re
from loguru import logger
import toytree
#from toytree.io.src.parse import WHITE_SPACE

logger = logger.bind(name="toytree")


class Network:
    """Child class of ToyTree that stores and draws network edges.

    A network object can be used to draw networks, to calculate or
    report statistics on a network, and to extract minor or major
    trees from a network (as ToyTree objects).

    Algorithm to traverse nodes of a network in order... could still
    use the same traversals that trace the major or minor tree to visit
    each node, but another traversal may be available that would also
    visit nodes connected by an admix edge sooner...
    """
    def __init__(self, network: Union[str, Path]):
        self.network = network
        # tree, admix = parse_network(net, disconnect=disconnect)
        # super().__init__(tree, *args, **kwargs)
        # self.admix = admix

    # def ... , disconnect: bool = True, **kwargs):

    def add_edge(self) -> toytree.ToyTree:
        """Return modified Network with a new degree-3 edge added."""

    def remove_edge(self) -> toytree.ToyTree:
        """Return modified Network with a degree-3 edge removed."""

    def get_major_tree(self) -> toytree.ToyTree:
        """Return the major tree..."""

    def get_admixture_edges(self) -> Tuple:
        """Return Tuple with list of admixture edges..."""

    def write(self):
        """Write network in extended format..."""

    def root(self):
        """raise rootMisMatchError if direction doesn't work..."""



# OLD IMPLEMENTATION FOR NetworkToMajorTree. NOW IN src/parse_network.py
# class NetworkToMajorTree:
#     """Class with functions to parse (ToyTree, admix_edges) from net.

#     ...
#     """
#     def __init__(self, net: Union[str, Path], disconnect=True):
#         self.net: str = self._get_network_as_newick_str(net)
#         """: Input network as newick string (parsed from str or Path)."""
#         self.tree: toytree.ToyTree = None
#         """: The major tree."""
#         self.admix: List[Tuple] = []
#         """: List of admixture edges."""
#         self.disconnect: bool = disconnect
#         """: ..."""

#     def _get_network_as_newick_str(self, net: Union[str, Path]) -> str:
#         """Parse network newick from file or '[network];...details' string."""
#         # if net is a file then read the first line
#         if ";" not in net:
#             if not net.exists():
#                 raise IOError("input type is not a file or newick (no ; ending).")
#             with open(net, 'rt', encoding="utf-8") as infile:
#                 net = infile.readline()
#         else:
#             # trim off loglik and anything after it (TODO: keep loglik)
#             if ";" in net: 
#                 net = net.split(";")[0] + ';'
#         # strip whitespace
#         net = WHITE_SPACE.sub("", net)
#         return net

#     def _set_tree_with_admix_as_node_labels(self) -> toytree.ToyTree:
#         """Return a ToyTree parsed from network newick with gamma nodes.

#         Admix nodes are indicated by #[name]:[dist]::[gamma] where dist
#         will be absent for tip nodes but present for internal nodes. We
#         modify this to be #[name-with-gamma]:dist so we can still use
#         the same tree parser and then modify the tree afterwards to
#         extract the major tree and minor admix edges.
#         """
#         string = self.net
#         # tip nodes replace L:::G -> L-G
#         re0 = re.compile(r"(#\w+):::([\d.]+)")
#         # internal nodes replace L:D::G -> L-G:D
#         re1 = re.compile(r"(#\w+):([\d.]+)::([\d.]+)")
#         string = re0.sub(r"\1-gamma-\2", string)
#         string = re1.sub(r"\1-gamma-\3:\2", string)
#         self.tree = toytree.tree(string)

#     def _pseudo_unroot(self) -> toytree.ToyTree:
#         """Return unrooted tree re-ordered for display purposes.

#         Tree is rooted tree on an edge that does not split a ...
#         """
#         # store a list of all nodes the tree *could* be rooted on.
#         opt = self.tree.get_nodes()
#         for node in self.tree.traverse("postorder"):
#             if node.name.startswith("#H"):
#                 for desc in node.get_descendants():
#                     if desc in opt:
#                         opt.remove(desc)
#         logger.trace(f"can root on {opt}")
#         return self.tree.root(opt[0]).mod.ladderize().mod.unroot()

#     def get_major_tree_and_admix_edges(self) -> Tuple:
#         """..."""
#         self._set_tree_with_admix_as_node_labels()
#         self.tree._draw_browser(ts='s', layout='unr', width=700, node_labels="name", node_colors="pink")
#         self.tree = self._pseudo_unroot()
#         self.tree._draw_browser(ts='s', width=500, node_labels="idx", node_colors="pink")

#         # restart traversal after each pair of nodes is fixed.
#         while 1:
#             try:
#                 hnodes = self.tree.get_nodes("#H*", regex=True)
#                 hnodes = sorted(hnodes, key=lambda x: x.name)
#             except ValueError:
#                 self.tree.mod.remove_unary_nodes(inplace=True)
#                 logger.debug(f"finished: {self.admix}")
#                 return self.tree, self.admix

#             # order as [dest, src] i.e., [node-label, tip-node] and get gamma
#             if hnodes[0].is_leaf():
#                 hnodes = [hnodes[1], hnodes[0]]
#             if "-gamma-" in hnodes[1].name:
#                 prop = hnodes[1].name.rsplit("-gamma-", 1)[1]
#             else:
#                 prop = 0.5
#             logger.trace(f"hnodes: {hnodes}")

#             # remove the tip node representing a src
#             logger.debug(f"deleting minor node: {hnodes[1].name} {hnodes[1].idx}")
#             hnodes[1]._delete()
#             desc1 = tuple(hnodes[1]._up.get_leaf_names())
#             desc1 = [i for i in desc1 if not i.startswith("#H")]
#             logger.trace(f"desc1: {desc1}")

#             # remove internal node representing admix destination
#             logger.debug(f"deleting major node: {hnodes[0].name} {hnodes[0].idx}")
#             desc0 = tuple(hnodes[0].get_leaf_names())
#             desc0 = [i for i in desc0 if not i.startswith("#H")]
#             hnodes[0]._delete()
#             logger.trace(f"desc0: {desc0}")

#             self.tree._update()
#             self.admix.append((desc0, desc1, 0.5, {}, f"{float(prop):.3g}"))'''


# def parse_network(net: Union[str, Path], disconnect=True):
#     """Parse a network file to extract a major topology and admix dict.

#     This leaves the hybrid nodes in the tree and labels each with
#     .name="H{int}" and .gamma={float}.
#     """
#     # if net is a file then read the first line
#     if ";" not in net:
#         if not net.exists():
#             raise IOError("input type is not a file or newick (no ; ending).")
#         with open(net, 'rt', encoding="utf-8") as infile:
#             net = infile.readline()
#     else:
#         # trim off loglik and anything after it (TODO: keep loglik)
#         if ";" in net:
#             net = net.split(";")[0] + ';'

#     # sub :xxx:: to be ::: b/c I don't care about admix edge bls FOR NOW.
#     net = re.sub(r":\d.\w*::", ":::", net)
#     print(net)

#     # change H nodes to be internal node labels rather than tip nodes.
#     while ",#" in net:
#         pre, post = net.split(",#", 1)
#         npre, npost = post.split(")", 1)
#         newpre = npre.split(":")[0] + "-" + npre.split(":")[-1]
#         net = pre + ")#" + newpre + npost
#     net = net.replace(":::", "-")
#     print(net)

#     # parse cleaned newick and set empty gamma on all nodes
#     net = toytree.tree(net)#, tree_format=1)
#     print(net.get_node_data())

#     # store admix data
#     admix = {}

#     # if not rooted choose any non-H root
#     if not net.is_rooted():
#         net = net.root(
#             [i for i in net.get_tip_labels() if not i.startswith("#H")][0]
#         )

#     # Traverse tree to find hybrid nodes. If a hybrid node is labeled as a
#     # distinct branch in the tree then it is dropped from the tree and
#     for node in net.traverse("postorder"):

#         # find hybrid nodes as internal nchild=1, or external with H in name
#         if (len(node.children) == 1) or node.name.startswith("#H"):

#             # assign name and gamma to hybrid nodes
#             print(node.name)
#             aname, aprop = node.name.split("-")
#             aname = aname.lstrip("#")
#             node.name = aname

#             # assign hybrid to closest nodes up and down from edge
#             # node.children[0].hybrid = int(aname[1:])
#             # node.gamma = round(float(aprop), 3)
#             # node.up.hybrid = int(aname[1:])

#             # if root is a hybrid edge (ugh)
#             if node.up is None:
#                 small, big = sorted(node.children, key=len)
#                 root = toytree.Node(name='root')
#                 node.children = [small]
#                 small.up = node
#                 node.up = root
#                 big.up = root
#                 root.children = [node, big]
#                 net.treenode = root

#             # disconnect node by connecting children to parent
#             if disconnect:

#                 # if tip is a hybrid
#                 if not node.children:
#                     # get sister node
#                     sister = [i for i in node.up.children if i != node][0]

#                     # connect sister to gparent
#                     sister.up = node.up.up
#                     node.up.up._remove_child(node.up)
#                     node.up.up._add_child(sister)

#                 # if hybrid is internal
#                 else:
#                     node.up._remove_child(node)
#                     for child in node.children:
#                         child._up = node.up
#                         node.up._add_child(child)

#             # store admix data by descendants but remove hybrid tips
#             desc = node.get_leaf_names()
#             if aname in desc:
#                 desc = [i for i in node.up.get_leaf_names() if i != aname]
#             desc = [i for i in desc if not i.startswith("#H")]

#             # put this node into admix
#             if aname not in admix:
#                 admix[aname] = (desc, aprop)

#             # matching edge in admix, no arrange into correct order by minor
#             else:
#                 # this is the minor edge
#                 if aprop < admix[aname][1]:
#                     admix[aname] = (
#                         admix[aname][0],
#                         desc,
#                         0.5,
#                         {},
#                         str(round(float(aprop), 3)),
#                     )

#                 # this is the major edge
#                 else:
#                     admix[aname] = (
#                         desc,
#                         admix[aname][0],
#                         0.5,
#                         {},
#                         str(round(float(admix[aname][1]), 3)),
#                     )

#     # update coords needed if node disconnection is turned back on.
#     # net._coords.update()
#     net._update()
#     net = net.mod.ladderize()
#     return net, admix

# def test3A():
#     net = "(((r0,#H1:::0.9),r2),(r3,(r4,#H1:::0.1)));"
#     return parse_network(net)

# def test3A():
#     """No gamma values."""
#     net = "(((r0,#H1),r2),(((r3,r4))#H1));"
#     parse = NetworkToMajorTree(net)
#     tree, admix = parse.get_major_tree_and_admix_edges()
#     return tree, admix

# def test3B():
#     """Major tree has higher gamma value.
#     >>> ((r0,r2),(r3,r4));  [(r0,(r3, r4), 0.5)]
#     """
#     # net = "(((r0,r1),#HX),(((r2,r3))#HX,r4),r5);"
#     # net = "(C,D,((O,(E,#H7:::0.196):0.314):0.664,(((A1,A2))#H7:::0.804,B):10.0):10.0);"
#     # net = "(C,D,((O,(E,#H7)),(B,(A)#H7)));"
#     # net = "(C,D,((O,(E,#H7:::0.49)),(B,(A)#H7:::0.51)));"
#     # net = "(C,D,((O,(E,#H7:::0.51)),(B,(A)#H7:::0.49)));"
#     net = "((r5,(r4,(r3,((r2,(r0,r1):4.268188213461936):2.9023407392778653)#H9:10.0::0.7093477022612464):3.2038798063864555):10.0):10.0,r6,(r7,#H9:0.0::0.2906522977387535):10.0);"
#     # net = "(C,D,((O,(E,#H7:::0.195):0.313):0.664,(B,(A)#H7:::0.804):10.0):10.0);"
#     parse = NetworkToMajorTree(net)
#     tree, admix = parse.get_major_tree_and_admix_edges()
#     return tree, admix

# def test3C():
#     """...
#     >>> ((r0,r2),(r3,r4));  [((r3, r4),r0, 0.5)]
#     """
#     NET = "(((r0,#H1:::0.9),r2),(r3,r4)#H1:::0.1);"

# def test_am_2():
#     net = """(quitensis,caudatus,((hybridus,((cruentus,((((retroflexus,wrightii):3.177882128684601,powellii):1.0960348982483923,acutilobus):1.196100087172878,((((((((tuberculatus,arenicola):0.6109501450803072,pumilus):0.4760189778806285,cannabinus):2.412166114595602,(((albus,blitoides):3.721391761165372,(viridis,(blitum,tricolor):1.0147162921429302):3.1242417846101658):0.8709001073072796,(fimbriatus)#H24:::0.5252749295119973):1.8571548128366557):2.707582965691606,#H24:::0.4747250704880027):1.5826753575568118,(((palmeri,watsonii):0.9398829701307351,spinosus):1.724020571200692)#H26:5.622450726846983::0.7964515993969742):3.198826088104374,(dubius,#H26:0.0005151894132288451::0.20354840060302581):9.999954155584865):0.6322904226943913)#H25:0.49722570736992866::0.7233188613740416):2.5040114155558104):0.47630091121000095,hypochondriacus):0.754444529685924):0.7024345634254702,#H25:2.415867402213081::0.2766811386259585):3.554563433506071);"""
#     parse = NetworkToMajorTree(net)
#     tree, admix = parse.get_major_tree_and_admix_edges()
#     return tree, admix
#     # parse = NetworkToMajorTree(NET)
#     # return parse
#     # return parse_network(NET, disconnect=0)

# def test_interactive_3():
#     NET = "(r1,(r2,(r3,(r4,#H6):10.0):1.949048825686914):10.0,(r0)#H6);"
#     NET = "((r3,(r2,(r1, #HX))),(r0)#HX);"
#     parse = NetworkParser(NET)
#     # parse_network(NET)





#!/usr/bin/env python

"""
Discrete State Markov Model Simulator. 

Some code from ipcoal (McKenzie and Eaton 2020).
"""

from typing import Optional, Union, List, Any
import pandas as pd
import numpy as np
import scipy.linalg


class QMatrixGenerator:
    """Generate a Q matrix under a number of models (ER, SYM, ARD).

    Returns a Q matrix under a model (ER, SYM, ARD) for n states 
    given user input for number of states, state_frequencies, and
    for the case of complex models (SYM with >2 states, or ARD), the
    relative rates as a matrix.

    Parameters
    ----------
    nstates: int
        The number of discrete Markov states.
    state_frequencies: 
        Equilibrium frequencies of states 0-n in order (should sum
        to one. 
    relative_rates: numpy.ndarray
        The relative transition rates between states as an array
        of size (nstates x nstates). Only relative differences among
        rates are relevant (will be auto-scaled). Values on the 
        diagonal are ignored. 

    Examples
    --------
    >>> # Generate an ER (equal-rates) normalized Q-matrix
    >>> qgen = DiscreteMarkovModelQMatrix(nstates=3)
    >>> qmatrix = qgen.get_rate_matrix()

    >>> # Generate an SYM (symmetric) normalized Q-matrix
    >>> qgen = DiscreteMarkovModelQMatrix(3, [0.1, 0.2, 0.3])
    >>> qmatrix = qgen.get_rate_matrix()

    >>> # Generate an ARD (all-rates-different) normalized Q-matrix
    >>> qgen = DiscreteMarkovModelQMatrix(3, [0.1, 0.2, 0.3])
    >>> qmatrix = qgen.get_rate_matrix()
    """
    def __init__(
        self,
        nstates: int=2,
        state_frequencies: Optional[Union[List,np.ndarray]]=None,
        ):
        assert nstates > 1, "discrete model must have >1 state"
        self.nstates = nstates
        self.state_frequencies = (
            state_frequencies if state_frequencies is not None 
            else np.repeat(1 / nstates, nstates)
        )

    def _get_relative_rates_er(self):
        """Return relative rates under the equal-rates model"""
        return np.repeat(1, self.nstates)

    def _get_relative_rates_sym(self, rates):
        """Return relative rates under the symmetric model"""
        nparams = (self.nstates * (self.nstates - 1)) / 2
        assert len(rates) == nparams, (
            f"the symmetric model with {self.nstates} should have "
            f"{nparams} rate parameters.")
        return np.repeat(1, self.nstates)

    def _get_non_normalized_qmatrix(self) -> np.ndarray:
        """Return the non-normalized instantaneous rate matrix.

        This matrix represents the nstates and their state frequencies
        before being normalized by a matrix scaling factor that will
        account for differences in state frequencies.

        Note
        ----
        This is not the Q-matrix you are looking for.

        Examples
        --------
        >>> qgen = DiscreteMarkovModelQMatrix(nstates=3)
        >>> qmatrix = qgen.get_non_normalized_qmatrix()
        [[-0.66666667  0.33333333  0.33333333]
         [ 0.33333333 -0.66666667  0.33333333]
         [ 0.33333333  0.33333333 -0.66666667]]

        >>> qgen = DiscreteMarkovModelQMatrix(3, [0.2, 0.4, 0.6])
        >>> qmatrix = qgen.get_non_normalized_qmatrix()
        [[-1.   0.4  0.6]
         [ 0.2 -0.8  0.6]
         [ 0.2  0.4 -0.6]]
        """
        # make non-normalized matrix (source sink)
        sfreqs = np.matrix(self.state_frequencies)
        ones = np.matrix(np.ones(sfreqs.size)).T
        non_normal_qmat = np.array(ones * sfreqs)
        for idx in range(ones.size):
            rowsum = non_normal_qmat[idx].sum()
            non_normal_qmat[idx, idx] = -(rowsum - non_normal_qmat[idx, idx])
        return non_normal_qmat

    def get_adjusted_rate_matrix(self, scalar: float=1.) -> np.ndarray:
        """Return the adjusted instantaneous rate matrix (Q).

        This is the Q-matrix that should be used in phylogenetic 
        comparative methods. It represents the expected instantaneous
        transition rates between nstates. It is calculated from the
        state frequencies ...

        Parameters
        ----------
        scalar: float
            Multiply the Q-matrix by a scalar to represent transition
            rates per unit of the scalar.

        Examples
        --------
        >>> qgen = DiscreteMarkovModelQMatrix(nstates=3)
        >>> qmatrix = qgen.get_adjusted_rate_matrix()
        [[-1.   0.5  0.5]
         [ 0.5 -1.   0.5]
         [ 0.5  0.5 -1. ]]

        >>> qgen = DiscreteMarkovModelQMatrix(3, [0.2, 0.4, 0.6])
        >>> qmatrix = qgen.get_adjusted_rate_matrix()
        [[-1.13636364  0.45454545  0.68181818]
         [ 0.22727273 -0.90909091  0.68181818]
         [ 0.22727273  0.45454545 -0.68181818]]
        """
        # get the non-normalized rate matrix
        non_normal_qmat = self._get_non_normalized_qmatrix()

        # get full matrix scaling factor
        scaling = [
            non_normal_qmat[i][i] * self.state_frequencies[i]
            for i in range(self.nstates)
        ]
        scaling = -1 / np.sum(scaling)

        # get adjusted rate matrix
        qmatrix = non_normal_qmat * scaling * scalar
        return qmatrix


class DiscreteMarkovModelSim:
    """Simulate a discrete n-state character on a phylogenetic tree.

    Input Q matrix (instantaneous transition rate matrix) can be 
    estimated from tip states using DiscreteMarkovModelFit.
    
    Parameters
    ----------
    tree
        ...
    qmatrix
        ...
    nsims
        ...

    Examples
    --------
    >>> import toytree
    >>> tree = toytree.rtree.unittree(10)
    >>> qmatrix = [[-0.7, 0.3], [0.5, -0.5]]
    >>> data = tree.pcm.simulate_discrete_markov_model(qmatrix, nsims=100)

    See also:
    ---------
    ...

    References:
    -----------
    ...
    """
    def __init__(
        self, 
        tree: 'toytree.ToyTree',
        qmatrix: Union[np.ndarray, pd.DataFrame],
        nsims: int = 1,
        ancestral_state: Optional[Any] = None,
        seed: int = None,
        ):

        # store input args
        self.rng = np.random.default_rng(seed)
        self.nsims = nsims
        self.tree = tree
        self.qmatrix = np.array(qmatrix)
        self.nstates = self.qmatrix.shape[0]
        self.ancestral_state = (
            ancestral_state if ancestral_state is not None
            else self.rng.choice(range(self.nstates), size=nsims)
        )

        # arg type checks
        assert self.nstates >= 2, "discrete model must have >1 states"
        assert self.ancestral_state.size == self.nsims, (
            "ancestral_state must be length nsims (set for every sim")
        assert np.allclose(self.qmatrix.sum(axis=1), 0), (
            f"rows of the Q matrix must sum to zero: {self.qmatrix.sum(axis=1)}")

    def _get_state_array(self) -> np.ndarray:
        """
        Rates are returned in units of changes per unit branch length, 
        and so should be interpreted in the context of whether blens
        are in units of years, mya, substitutions, etc.
        """
        data = np.zeros((self.nsims, self.tree.ntips), dtype=int)
        for sim in range(self.nsims):
            trait = {self.tree.treenode.idx: self.ancestral_state[sim]}
            for node in self.tree.treenode.traverse():
                if not node.is_root():
                    # probability of change over this length of branch
                    prob_mat = scipy.linalg.expm(node.dist * self.qmatrix)
                    # sample from probability matrix
                    trait[node.idx] = np.argmax(
                        np.random.multinomial(1, prob_mat[trait[node.up.idx]]))
            data[sim] = [trait[i] for i in range(self.tree.ntips)]
        return data

    def _get_state_tree(self) -> List['toytree.ToyTree']:
        """
        Rates are returned in units of changes per unit branch length, 
        and so should be interpreted in the context of whether blens
        are in units of years, mya, substitutions, etc.
        """
        data = []
        for sim in range(self.nsims):
            trait = {self.tree.treenode.idx: self.ancestral_state[sim]}
            for node in self.tree.treenode.traverse():
                if not node.is_root():
                    # probability of change over this length of branch
                    prob_mat = scipy.linalg.expm(node.dist * self.qmatrix)
                    # sample from probability matrix
                    trait[node.idx] = np.argmax(
                        np.random.multinomial(1, prob_mat[trait[node.up.idx]]))
            ntre = self.tree.set_node_data(feature='discrete', mapping=trait)
            data.append(ntre)
        return data


if __name__ == "__main__":

    import toytree
    TREE = toytree.rtree.unittree(10, treeheight=100, seed=123)

    # use a hand-made matrix
    ALPHA = 0.001
    BETA = 0.002
    QMATRIX = np.array([
        [-ALPHA, ALPHA],
        [BETA, -BETA],
    ])

    # equal rates model
    QGEN = QMatrixGenerator(nstates=3)
    print(QGEN.get_adjusted_rate_matrix(1))    
    # print(QGEN._get_non_normalized_qmatrix())

    QGEN = QMatrixGenerator(nstates=3, state_frequencies=[0.2, 0.4, 0.6])
    # print(QGEN._get_non_normalized_qmatrix())
    print(QGEN.get_adjusted_rate_matrix(1))

    print(QGEN.get_adjusted_rate_matrix(0.5))

    # or generate a symmetric matrix
    QGEN = QMatrixGenerator(nstates=3, state_frequencies=[0.25, 0.2, 0.15])
    print(QGEN._get_non_normalized_qmatrix())
    print(QGEN.get_adjusted_rate_matrix())

    QMATRIX = QGEN.get_adjusted_rate_matrix(5e-3)

    # show matrix and simulate
    print(QMATRIX)
    print(toytree.pcm.simulate_discrete_markov_data(TREE, QMATRIX, nsims=3))
    # print(toytree.pcm.simulate_discrete_markov_data_on_tree(TREE, QMATRIX, nsims=1))



#!/usr/bin/env python

"""Annotation methods for adding node markers to tree drawings.

Examples
--------
...
"""

from typing import Tuple, Sequence, Mapping, Any, Union, TypeVar
import numpy as np

from toytree.core import ToyTree, Cartesian, Mark
from toytree.color import ToyColor
from toytree.core.apis import add_subpackage_method, AnnotationAPI
from toytree.style.src.validate_utils import substyle_dict_to_css_dict
from toytree.annotate.src.checks import get_last_toytree_mark, assert_tree_matches_mark
from toytree.style.src.validate_data import (
    validate_colors,
    validate_numeric,
    validate_mask,
    validate_labels,
)
from toytree.style.src.validate_node_labels import validate_node_labels_style

Color = TypeVar("Color", str, tuple, np.ndarray)
__all__ = ["add_tip_labels"]

@add_subpackage_method(AnnotationAPI)
def add_tip_labels(
    tree: ToyTree,
    axes: Cartesian,
    labels: Union[str, Sequence[str]] = "name",
    color: Union[Color, Sequence[Color]] = None,
    opacity: Union[float, Sequence[float]] = 1.0,
    font_size: Union[int, None] = 12,
    angle: Union[int, Sequence[int]] = 0,
    mask: Union[np.ndarray, Tuple[int, int, int], None] = None,
    xshift: int = 0,
    yshift: int = 0,
    style: Mapping[str, Any] = None,
) -> Mark:
    """Return a toyplot Mark of tip labels added to a tree drawing.

    This adds text labels to leaf Nodes on the selected tree (or 
    last tree) drawn on the Cartesian axes.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    labels: str or Sequence[str]
        A sequence of labels as strings in Node idx traversal order.
        The length must be equal to nnodes or nnodes - 1. Use '' to
        not add a label to some edges. You can also enter a single str
        as the name of a feature on the tree to extract data as labels,
        e.g., "idx" will show int idx labels for nodes.
    color: str, tuple, array or Sequence
        A single color or Sequence of colors for node labels.
    opacity: float or Sequence[float]
        A single opacity or Sequence of opacities for node labels.
    font_size: float
        Font size in px. Overrides 'font-size' setting in style dict.
    angle: int or Sequence[int]
        A single angle applied to all labels, or Sequence of angles.
    mask: np.array or None
        A boolean array of len nnodes or nnodes - 1 where True masks
        an node from being shown and False shows the node. None or False
        shows all nodes. A tuple of 3 booleans can be entered as a
        shortcut to (show_tips, show_internal, show_root).
    xshift: int
        Shift label horizontally by px units (+=right, -=left).
    yshift: int
        Shift label vertically by px units (+=down, -=up).
    style: dict
        Style dict. See `tree.style.node_labels_style` for options.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(6, seed=123)
    >>> canvas, axes, m0 = tree.draw()
    >>> m1 = tree.annotate.add_tip_labels(
    >>>     axes,
    >>>     labels="name",
    >>>     color="blue",
    >>>     font_size=15,
    >>> )
    """
    # get mark for coordinates on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # mask some edges
    if mask is None:
        mask = tree.get_node_mask(show_tips=True)[:tree.ntips]
    else:
        if mask.size == tree.ntips:
            # add False for all internal nodes to make the mask = nnodes
            mask = np.concatenate(mask, np.full(tree.nnodes - tree.ntips), False)
        if mask.size == tree.nnodes:
            mask = validate_mask(tree, style={"node_mask": mask})[:tree.ntips]
        else:
            raise ValueError("mask should be a boolean array of size ntips or nnodes")
    coords = mark.ntable[:tree.ntips][mask]

    # check length and type of labels
    labels = validate_labels(
        tree, key="labels", size=tree.ntips, style={"labels": labels})[mask]

    # set styles on top of defaults
    style = {} if style is None else style
    style["text-anchor"] = style.get("text-anchor", "start")
    style = validate_node_labels_style(tree, style=style, **style)
    style = substyle_dict_to_css_dict(style.__dict__)

    # override font size
    if font_size:
        style["font-size"] = font_size

    # update node colors setting; sets to None if only one color.
    node_colors, fill_color = validate_colors(
        tree, key="color", size=tree.ntips, style={"color": color})

    # if fill_color then set to node_style.fill since node_colors = None
    if node_colors is None:
        if fill_color:
            style["fill"] = ToyColor(fill_color)  # overrides node_style.fill
        else:
            pass  # node_style.fill overrides
    else:
        node_colors = node_colors[mask]
        style.pop("fill")

    # ...
    opacity = validate_numeric(
        tree, key="opacity", size=tree.ntips, style={"opacity": opacity})[mask]
    angle = validate_numeric(
        tree, key="angle", size=tree.ntips, style={"angle": angle})[mask]

    # expand xshift,yshift args in pixel units
    style['-toyplot-anchor-shift'] += xshift
    style['baseline-shift'] -= yshift

    # add text at Node positions + half length of dists.
    mark = axes.text(
        coords[:, 0],
        coords[:, 1],
        labels,
        color=node_colors,
        opacity=opacity,
        angle=angle,
        style=style,
        annotation=True,
    )
    return mark


if __name__ == "__main__":
    pass



#!/usr/bin/env python

r"""Class for rooting and unrooting of trees.

Rooting and unrooting trees turns out to be quite a bit more
complicated than you might expect. The difficult part is keeping
track of Node features that in practice are coded as attributes to
Nodes, but in reality are sometimes features of *edges*, such as
'support' or 'dist', and thus need to be re-assigned when an parent
child relationships change.

Re-rooting a rooted tree:
               o                            x
              / \                          / \
             1   2         root('n')      n   u
                / \          -->             / \
               .   u                        2   .
                  / \                      / \
                 .   n                    1   .

Rooting an unrooted tree:
                                            x
                                           / \
              _ 2 _        root('n')      n   u
             |  |  |         -->             / \
             1  .  u                        2   .
                  / \                      / \
                 .   n                    1   .

References
----------
- Czech, L., Huerta-Cepas, J. and Stamatakis, A. (2017) A critical
  review on the use of support values in tree viewers and
  bioinformatics toolkits. Molecular Biology and Evolution, 34,
  1535–1542. doi: 10.1093/molbev/msx055

Examples
--------
>>> # Test Case from Czech et al. 2017 (edge features)
>>> NEWICK = "((C,D)1,(A,(B,X)3)2,E)R;"
>>> tree = toytree.tree(NEWICK)
>>>
>>> # set a color feature to Nodes and 'ecolor' and 'ncolor'
>>> colors = {'1': 'red', '2': 'green', '3': 'orange'}
>>> tree.set_node_data("ecolor", colors, inplace=True)
>>> tree.set_node_data("ncolor", colors, inplace=True)
>>>
>>> # draw the trees
>>> kwargs = {
>>>     'layout': 'd',
>>>     'use_edge_lengths': False,
>>>     'node_sizes': 10,
>>>     'node_labels': 'name',
>>>     'node_labels_style': {
>>>         'font-size': 20,
>>>         'baseline-shift': 10,
>>>         '-toyplot-anchor-shift': 10,
>>>     }}
>>> tree.draw(
>>>     node_colors=tree.get_node_data('ncolor', missing='black'),
>>>     edge_colors=tree.get_node_data('ecolor', missing='black'),
>>>     **kwargs,
>>> )
>>>
>>> # re-root, treating 'ecolor' but not 'ncolor' as edge feat.
>>> rtree = tree.root("X", edge_features=["ecolor"])
>>> rtree.draw(
>>>     node_colors=rtree.get_node_data('ncolor', missing='black'),
>>>     edge_colors=rtree.get_node_data('ecolor', missing='black'),
>>>     **kwargs,
>>> );
"""
from typing import Optional, Sequence, TypeVar, Set, Union
from loguru import logger
import numpy as np
from toytree.core.node import Node
from toytree.core.tree import ToyTree
from toytree.core.apis import (
    TreeModAPI, add_subpackage_method, add_toytree_method
)
from toytree.utils import ToytreeError, NON_MONOPHYLETIC_OUTGROUP

logger = logger.bind(name="toytree")
Query = TypeVar("Query", int, str, Node)

__all__ = ["root", "unroot"]


class Rooter:
    """Root ToyTree on Node query selection.

    See `toytree.mod.root` docstring for details. This class is for
    internal use, not user-facing.
    """
    def __init__(
        self,
        tree: ToyTree,
        *query: Query,
        root_dist: Optional[float] = None,
        edge_features: Optional[Sequence[str]] = None,
        inplace: bool = False,
    ):
        self.tree = tree
        self.query = query
        self.root_dist = root_dist
        self.inplace = inplace
        self.node: Node = self._get_edge_to_split()
        """: The Node below the edge that will be split. Found from query."""
        self.edge_features: Set[str] = self._get_edge_features(edge_features)
        """: Features that should be re-polarized on rooting (e.g., support)"""

    def _get_edge_features(self, edge_features: Optional[Union[str, Sequence[str]]]) -> Set[str]:
        """Return the features associated with edges instead of nodes.

        This requires dist and support to be edge features. It also adds
        any features in self.tree.edge_features, and also adds any that
        were entered as args to the Rooter class.
        """
        default = {"_dist", "support"}
        disallowed = {"dist", "idx", "up", "children"}

        # start with tree's designated edge features
        efeatures = self.tree.edge_features
        # add user entered single feature
        if isinstance(edge_features, str):
            efeatures = self.tree.edge_features | {edge_features}
        # add user entered sequence of features
        elif isinstance(edge_features, (tuple, list, set)):
            efeatures = efeatures | set(edge_features)
        # add default features
        efeatures = efeatures | default
        # remove unallowed features
        efeatures = efeatures - disallowed
        logger.debug(f"edge_features: {efeatures}")
        return efeatures
        # return (default | set(edge_features)) - disallowed

    def _get_edge_to_split(self) -> Node:
        """Return the Node below new root edge from input query.

        Get MRCA Node of the input selection. If it is the root, then
        get the MRCA of the inverse selection. If that group is not
        monophyletic then raise exception for bad selection.
        """
        # get the query as Set[Node]; remove root if present; return if None
        nodes = set(self.tree.get_nodes(*self.query))
        nodes -= {self.tree.treenode}
        if not nodes:
            return self.tree.treenode

        # get only tips and get their mrca
        tips = set.union(*(set(i.get_leaves()) for i in nodes))
        mrca = self.tree.get_mrca_node(*nodes)

        # check monophyly of user query; try reciprocal tip set; then raise
        try:
            # the mrca clade may be non-monophyletic
            if not mrca.is_root():
                if any(node not in tips for node in mrca.iter_leaves()):
                    tips = str(sorted(i.name for i in tips))
                    raise ToytreeError(NON_MONOPHYLETIC_OUTGROUP.format(tips))

            # root mrca. Reciprocal tip set may be monophyletic, all tips, or non-monophyletic
            else:
                all_tips = set(self.tree[:self.tree.ntips])
                tips = all_tips - tips
                if not tips:
                    tips = all_tips
                mrca = self.tree.get_mrca_node(*tips)
                if any(node not in tips for node in mrca.iter_leaves()):
                    tips = str(sorted(i.name for i in tips))
                    raise ToytreeError(NON_MONOPHYLETIC_OUTGROUP.format(tips))
        except ToytreeError as exc:
            logger.error(exc)
            raise exc
        return mrca

    def run(self):
        """Return ToyTree rooted on the input selection."""
        # get tree and node copies
        if not self.inplace:
            self.tree = self.tree.copy()
            self.node = self.tree[self.node.idx]

        # tree is currently rooted, return it or unroot it.
        if self.tree.is_rooted():
            # return same if rooting on current rooting.
            if self.tree.treenode in [self.node, self.node.up]:
                logger.debug("tree is already rooted this way.")
                return self.tree
            self.tree.unroot(inplace=True)

        # tree is currently unrooted, raise exception if no outgroup.
        else:
            if self.node.is_root():
                msg = (
                    "Cannot root unrooted tree on the pseudo-root, it has no "
                    "edge. Select a valid outgroup")
                logger.error(msg)
                raise ToytreeError(msg)

        self._insert_root_node()
        self.tree._update()
        return self.tree

    def _insert_root_node(self):
        r"""Insert a new node to break an edge to create root.

        When inserting a new Node it is given nan support value.

        Important Nodes are:
            - oldroot: o
            - oldroot.children: 1,2
            - node: n
            - node.up: u
            - newroot: x

                   o                         x         polar = [u, 2]
                  / \                       / \        flip = [1]
                 1   2       root('n')     n   u
                    / \         -->           / \
                   .   u                     2   .
                      / x                   / \
                     .   n                 1   .

        oldroot is removed, newroot is created. All Nodes that are
        ancestors of u on the original tree are re-polarized.
        """
        # store references to Nodes before their relationships change.
        # support = self._infer_max_support()
        # self.node.support = np.nan
        newroot = Node(name="root", support=np.nan, dist=0)
        edge = (self.node, self.node.up)
        odist = float(edge[0]._dist)
        # odist = sum([i._dist for i in self.node.up.children if i != self.node])
        # sumdist = sum([ndist, odist])
        # logger.warning(f"edge={edge}, ndist={ndist} odist={odist}")

        # nodes on path from node to the original root.
        path = (self.node,) + self.node.get_ancestors()  # [n, u, 2, o]

        # store edge features for re-polarizing edges in path.
        feats = {
            n: {i: getattr(n, i, None) for i in self.edge_features}
            for n in path
        }

        # polarize each node on path
        for idx in range(1, len(path)):           # [u, 2, o]    [3, 2, R]
            # connect this node to its new parent (previously child)
            node = path[idx]                      # u            R
            child = path[idx - 1]                 # n            2
            node._up = child                      # u -> n       R -> 2
            node._remove_child(child)             # n -x u       2 -x R

            # connect this node to its new child (previously parent)
            # if this node is not the former root node.
            if node != path[-1]:
                node._add_child(path[idx + 1])

            # transfer non-root edge features
            for feature in self.edge_features:
                value = feats[child][feature]
                if value is not None:
                    setattr(node, feature, value)
                else:
                    if hasattr(node, feature):
                        delattr(node, feature)

        # set support of mirror root child to nan
        # self.node.support = np.nan

        # set node and its parent as sisters, and children of newroot.
        newroot._children = edge             # n -> x, u -> x
        edge[0]._up = newroot
        edge[1]._up = newroot

        # position of newroot on edge, None=midpoint, ...
        if self.root_dist is None:
            edge[0]._dist = odist / 2.
            edge[1]._dist = odist / 2.
            # edge[0]._dist /= 2.
            # edge[1]._dist /= 2.
        else:
            logger.debug(f"sumdist={odist}, root_dist={self.root_dist}")
            if self.root_dist > odist:
                raise ValueError(
                    "`root_dist` arg (placement of root node on existing edge) "
                    f"cannot be greater than length of the edge: {odist}.")
            edge[0]._dist = self.root_dist
            edge[1]._dist = odist - self.root_dist
            # logger.warning(f"sumdist={sumdist}, root_dist={self.root_dist}")
            # logger.warning(f"0={edge[0]._dist}, 1={edge[1]._dist}")

        # update as ToyTree
        self.tree.treenode = newroot
        self.tree._update()
        return self.tree


@add_toytree_method(ToyTree)
@add_subpackage_method(TreeModAPI)
def root(
    tree: ToyTree,
    *query: Query,
    root_dist: Optional[float] = None,
    edge_features: Optional[Sequence[str]] = None,
    inplace: bool = False,
) -> ToyTree:
    r"""Return a ToyTree rooted on the edge above selected Node query.

    Manually root a tree on an outgrup by splitting an edge to insert
    a new root Node. The root Node is named "root" and has a np.nan
    support value and dist=0 unless otherwise set.

    Example of rooting an unrooted tree:
                                                x
                                               / \
                  _ 2 _        root('n')      n   u
                 |  |  |         -->             / \
                 1  .  u                        2   .
                      / \                      / \
                     .   n                    1   .

    Example of re-rooting a rooted tree:
                   x                            x
                  / \                          / \
                 1   2         root('n')      n   u
                    / \          -->             / \
                   .   u                        2   .
                      / \                      / \
                     .   n                    1   .

    Parameters
    ----------
    tree: ToyTree
        A rooted or unrooted ToyTree to (re-)root.
    *query: str, int, or Node
        One or more Node selectors, which can be Node objects, names,
        or int idx labels. If multiple are entered the MRCA node will
        be used as the base of the edge to split.
    root_dist: None or float
        The length (dist) along the root edge above the Node query
        where the new root edge should be placed. Default is None
        which will place root at the midpoint of the edge. A float
        can be entered, but will raise ToyTreeError if > len of edge.
    edge_features: Sequence[str]
        One or more Node features that should be treated as a feature
        of its edge, not the Node itself. On rooting, edge features
        are re-polarized, to apply to the correct Node. The 'dist'
        and 'support' features are always treated as edge features.
        Add additional edge features here. See docs for example.
    inplace: bool
        If True the original tree is modified and returned, otherwise
        a modified copy is returned.

    See Also
    --------
    - toytree.mod.root_on_midpoint
    - toytree.mod.root_on_minimal_ancestor_deviation

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=10, seed=123)
    >>> t1 = tree.root("r8", "r9")
    >>> t2 = tree.root("r8", "r9", root_dist=0.3)
    >>> toytree.mtree([t1, t2]).draw();
    """
    kwargs = dict(inplace=inplace, root_dist=root_dist, edge_features=edge_features)
    return Rooter(tree, *query, **kwargs).run()


@add_toytree_method(ToyTree)
@add_subpackage_method(TreeModAPI)
def unroot(tree: ToyTree, inplace: bool = False) -> ToyTree:
    """Return an unrooted tree by collapsing the root split.

    This will convert a binary split into a multifurcation with three
    or more children. This decreases the number of Nodes by one.

    Note
    ----
    The unrooting process is not destructive of information, you can
    re-root a tree on the same edge position as before to recover the
    same tree, and all edge lengths (dist values) are retained. Only
    the position of the root Node along the rooted edge is lost, which
    can be re-set when rooting using the root_dist argument.

    Parameters
    ----------
    inplace: bool
        If True modify and return original tree, else return a copy.
    """
    # fast: return current tree if already unrooted
    if not tree.is_rooted():
        return tree

    # get tree or copy to return
    tree = tree if inplace else tree.copy()
    rootnode = tree.treenode

    # just return current tree if the rootnode node is not binary
    if len(rootnode.children) != 2:
        return tree

    # find a child with children, checking first left then right.
    if not rootnode.children[0].is_leaf():
        newroot = rootnode.children[0]
        ochild = rootnode.children[1]
    elif not rootnode.children[1].is_leaf():
        newroot = rootnode.children[1]
        ochild = rootnode.children[0]
    else:
        raise ToytreeError("Cannot unroot a tree with only two leaves")

    # child becomes ochild's new parent
    ochild._up = newroot
    newroot._children += (ochild,)

    # other child's dist extends to include child->oldrootnode dist
    ochild._dist += newroot.dist
    newroot._dist = 0.

    # ochild->child edge inherits features from child->oldrootnode edge
    ochild.support = newroot.support
    newroot.support = np.nan

    # make child the new rootnode and remove old rootnode
    newroot._up = None

    tree.treenode = newroot
    del rootnode

    # update idxs and return
    tree._update()
    return tree


if __name__ == "__main__":

    # Example test: start with a simple balanced tree.
    import toytree
    toytree.set_log_level("WARNING")
    TREE = toytree.rtree.imbtree(ntips=10)
    print(TREE)

    # root on
    TREE.unroot().root('r0', 'r1', 'r2')

    # raise error b/c not monophyletic
    # TREE.unroot().root("r8", "r9", "r0")

    # c, a, m = unroot(TREE).draw()
    # _, a, m = root(TREE, 'r2')._draw_browser()

    # T0 = root(TREE, 'r2', root_dist=0.3)

    # c2, a, m = unroot(TREE)._draw_browser(axes=a)
    # print(unroot(TREE))

    # unroot the tree
    # TREE = TREE.unroot()

    # root the tree a clade with the first two samples
    # TREE = TREE.root(['r1', 'r2'])

    # re-root on a different clade, for last two samples
    # TREE = TREE.root(['r9', 'r10'])

    # check that dist and support values were properly retained.
    # ...


#!/usr/bin/env python

"""Get quad splits on a tree as (child1, child2, sister, up) tuples.

Methods
-------
Get sets of Nodes from the four subclades of each edge in a tree.
>>> tree._iter_quadripartition_sets()
# (({0}, {1}), ({2}, {3, 4})), ...

Get tuples of Nodes for each quartet induced by quadripartitions in a tree.
>>> tree.iter_quadripartitions()
# (((0,), (1,)), ((2,), (3, 4))), ...

Get tuples of Nodes for each quartet induced by quadripartitions in a tree.
>>> tree.iter_quadripartitions('idx', sort=True, collapse=True)
# ({'c'}, {'d'}, {'a', 'b'}, {'e', 'f'})
"""

from typing import TypeVar, Iterator, Tuple, Optional, Set, Callable, Sequence
import itertools
from loguru import logger
from toytree import Node, ToyTree
from toytree.core.apis import TreeEnumAPI, add_subpackage_method, add_toytree_method
# from toytree.utils import ToytreeError

logger = logger.bind(name="toytree")
Query = TypeVar("Query")
# Node = TypeVar("Node")

__all__ = [
    "_iter_quadripartition_sets",
    "iter_quadripartitions",
    # "iter_edge_quadripartition_sets",
]


@add_subpackage_method(TreeEnumAPI)
def _iter_quadripartition_sets(
    tree: ToyTree,
    feature: Optional[str] = "name",
    contract_partitions: bool = False,
    include_internal_nodes: bool = False,
) -> Iterator[Tuple[Tuple[Set, Set], Tuple[Set, Set]]]:
    """Yield a tuple of sets of tips for each quadripartition.

    Quadripartitions are yielded in Node idxorder traversal as a tuple
    of two tuples of two sets, e.g., `(({e0},{e1}), ({e2},{e3}))`,
    representing sets of (tip) Nodes descending from each of the four
    edges subtending a focal edge. In this example, the focal edge
    splits e0,e1 from e2,e3. Note the use of sets to indicate that the
    items are not sorted.

    Notes
    -----
    - The order of yielded quadripartition sets is in Node idx order.
    - The order of partitions within a quad is (child-left, child-right,
    sister, up) unless at root in an unrooted tree, then it is
    (child-left, child-right, sister-left, sister-right)

    Parameters
    ----------
    tree: ToyTree
        A tree to extract edge sets from.
    feature: str or None
        Feature to represent Nodes in returned sets. Default is 'name'
        to use Node name strings. None will return Node objects.
    contract_partitions: bool
        If True then each partition is contracted to show only the
        first Node closest to the edge.
    include_internal_nodes: bool
        If True then all nodes in each partition are shown, whether it
        is internal or a tip Node. If False then only tips are shown.
        Note: This option is overriden if contract_partitions=True in
        which case the closest Node is shown whether or not it is a tip.

    Examples
    --------
    >>> newick = "((a,b)X,((c,d)Y,e)Z)R;"
    >>> tree = toytree.tree(newick)
    >>> tree.draw(ts='r');

    >>> sorted(_iter_quadripartition_sets(tree))
    >>> # (({'a'}, {'b'}), ({'d', 'c'}, {'e'}))
    >>> # (({'c'}, {'d'}), ({'e'}, {'a', 'b'}))

    >>> kwargs = dict(include_internal_nodes=True)
    >>> sorted(_iter_quadripartition_sets(tree, **kwargs))
    >>> # ({'a'}, {'b'}), ({'c', 'Y', 'd'}, {'e'}))
    >>> # (({'c'}, {'d'}), ({'e'}, {'a', 'X', 'b'}))

    >>> kwargs = dict(contract_partitions=True)
    >>> sorted(iter_edge_quadripartition_sets(tree, **kwargs))
    >>> # (({'a'}, {'b'}), ({'Y'}, {'e'}))
    >>> # (({'c'}, {'d'}), ({'e'}, {'Z'}))
    """
    # tree = tree.unroot()
    cache = {}

    # build a cache of descendants for each Node.
    for node in tree[:-1]:

        # if tip simply store its info
        if node.is_leaf():
            cache[node] = {node}
            continue

        # get tips from each child, and store union to this node
        vals = [cache[i] for i in node.children]

        # cache only the closest Node (exclude_internal_nodes ignored)
        if contract_partitions:
            cache[node] = {node}

        # cache tip Nodes with or without internal Nodes
        else:
            # cache only the tip Nodes
            if not include_internal_nodes:
                cache[node] = set.union(*vals)
            # cache all nodes including internals
            else:
                cache[node] = set.union(*vals) | {node}

    # get set of all relevant nodes for set diffs below
    if include_internal_nodes:
        nodes = set(cache)
    else:
        nodes = set(tree.get_nodes()[:tree.ntips])

    # iterate over internal nodes/edges in the tree
    topnode = tree.nnodes - 2 if tree.is_rooted() else tree.nnodes - 1
    for node in tree[tree.ntips: topnode]:

        # get >= 2 sets of nodes for children below
        below = [cache[i] for i in node.children]

        # if up is root then choose (child, child, sister, sister)
        if node._up.is_root():
            above = []
            sisters = node.get_sisters()
            if len(sisters) > 1:
                for sis in sisters:
                    above.append(cache[sis])
            else:
                for sis_child in sisters[0]._children:
                    above.append(cache[sis_child])

            # contract...

            # create generator
            _below = itertools.combinations(below, 2)
            _above = itertools.combinations(above, 2)
            qiter = itertools.product(_below, _above)

        # else choose (child, child, sister, up)
        else:
            # get >= 1 sets of nodes above edge, sister to children.
            up_down = [cache[sis] for sis in node.get_sisters()]

            # get set nodes above edge
            up_up = [nodes - {node, node._up} - set.union(*up_down + below)]

            # unless contracting, then just the parent
            if contract_partitions:
                up_up = [{node._up}]

            # create generator
            _below = itertools.combinations(below, 2)
            _up_down = itertools.combinations(up_down, 1)
            _up_up = itertools.combinations(up_up, 1)
            qiter = itertools.product(_below, _up_down, _up_up)

        # expand qiter(2, 2) or qiter(2, 1, 1)
        for below, above, *other in qiter:
            if len(above) == 1:
                x1, x2 = below
                x3 = above[0]
                x4 = other[0][0]
            else:
                x1, x2 = below
                x3, x4 = above

            # require all to exclude root in unrooted tree: ()
            if feature is None:
                yield ((x1, x2), (x3, x4))
            else:
                yield (
                    tuple({getattr(j, feature) for j in i} for i in (x1, x2)),
                    tuple({getattr(j, feature) for j in i} for i in (x3, x4)),
                )


@add_subpackage_method(TreeEnumAPI)
def iter_quadripartitions(
    tree: ToyTree,
    feature: Optional[str] = "name",
    contract_partitions: bool = False,
    include_internal_nodes: bool = False,
    collapse: bool = False,
    type: Callable = set,
    sort: bool = False,
) -> Iterator[Tuple]:
    """Generator of quadripartitions (Nodes on four clades of a split)

    Quadripartitions represent the splits in a tree. Many algorithms
    compare tips (or internal Nodes) split into the four parts of a
    quadripartition to compute metrics on trees. This function aims to
    provide a flexible and fast framework for yielding quadripartitions
    in various formats.

    Notes
    -----
    - qparts are generated in Node idx traversal order.
    - By default qparts are formatted as a tuple of two tuples of sets,
    e.g., ((set1, set2), (set3, set4)), but can be optionally collapsed
    into the format (set1, set2, set3, set4) where the split in the
    middle is implicit.
    - The partitions can be returned as other types than sets using
    the `type` arg, in which case the items within a part will be
    sorted by names.
    - The order of the two tuples is (below, above) given the traversal
    order of edges, but can instead be sorted consistently to allow for
    comparisons by sorting by len and then names using the `sort` arg.

    Parameters
    ----------
    feature: str
        Feature to return to represent Nodes in partitions. Default is
        "name". None will return Node objects. Any other Node feature,
        such as "idx", is also supported. Note the feature arg does not
        affect the order in which partitions are sorted (see `sort`).
    include_internal_nodes: bool
        If True then internal Node names/features are included in
        addition to tip Nodes.
    contract_partitions: bool
        If True then clades are represented only by the Node (internal
        or tip) closest to the quadripartition edge.
    collapse: bool
        If True then the format is simplified from `((a,b),(x,y))`
        to `(a,b,x,y)` where the split ab|xy is still implicit.
    type: Callable
        The type of collection used to represent a partition. Default
        is `set` to return a tuple of sets, but another useful option
        is `tuple`, which returns a tuple of tuples. The latter
        collection can be converted into a set of quadripartitions.
    sort: bool
        If False, tuple clades are returned as (child, parent) order
        given the topology and rooting in Node idx order traversal. If
        sort=True, qpartitions are instead always sorted first by len,
        e.g., (fewer, longer) and if the same len, then next by the
        lowest alphanumeric tip name, e.g., ({'a', 'b'}, {'c', 'd'}).
        If the requested partition `type` is sortable (i.e., not a set)
        then items within a partition are also consistently sorted.

    Examples
    --------
    >>> tree = toytree.tree("(a,b,((c,d)CD,(e,f)EF)X)AB;")

    >>> # default: parts w/ tip names as (child, parent) in idx order
    >>> list(tree.enum.iter_quadripartitions())
    >>> # [(({'c'}, {'d'}), ({'e', 'f'}, {'b', 'a'}))
    >>> #  (({'e'}, {'f'}), ({'d', 'c'}, {'b', 'a'}))
    >>> #  (({'d', 'c'}, {'e', 'f'}), ({'a'}, {'b'}))]

    >>> # same order, but as int idx labels including internal Nodes
    >>> list(tree.enum.iter_quadripartitions(
    >>>     feature='idx',
    >>>     include_internal_nodes=True,
    >>> ))
    >>> # [(({2}, {3}), ({4, 5, 7}, {0, 1})),
    >>> #  (({4}, {5}), ({2, 3, 6}, {0, 1})),
    >>> #  (({2, 3, 6}, {4, 5, 7}), ({0}, {1}))]

    >>> # args to get consistently sorted qparts regardless of rooting
    >>> sorted(tree.enum.iter_quadripartitions(type=tuple, sort=True))
    >>> # ((('a',), ('b',)), (('c', 'd'), ('e', 'f')))
    >>> # ((('c',), ('d',)), (('a', 'b'), ('e', 'f')))
    >>> # ((('e',), ('f',)), (('a', 'b'), ('c', 'd')))

    >>> # example: easy comparison of consistently sorted sets
    >>> x = set(tree.root('a').enum.iter_quadripartitions(type=tuple, sort=True))
    >>> y = set(tree.root('e').enum.iter_quadripartitions(type=tuple, sort=True))
    >>> assert x == y
    """
    kwargs = dict(
        tree=tree,
        feature=feature,
        contract_partitions=contract_partitions,
        include_internal_nodes=include_internal_nodes,
    )

    # fastest approach returns bipart consistently as (child, parent)
    # and get feature from _iter_bipartition_sets if requested.
    # yield ({part1}, {part2})
    if (not sort) and (type == set):
        for below, other in _iter_quadripartition_sets(**kwargs):
            if collapse:
                yield below[0], below[1], other[0], other[1]
            else:
                yield below, other

    # slower approach sorts parts alphanumerically by tip Node names.
    # sort requires first getting objects as Nodes so we can access
    # the Name feature regarldess of what the returned feature will be.
    else:
        # yield as type(*Nodes) or as type(feature(*Nodes))
        if feature is not None:
            def tformat(node_tuple):
                return type(getattr(i, feature) for i in node_tuple)
        else:
            tformat = type

        # iterate over bipartitions and order items within
        kwargs['feature'] = None
        for below, other in _iter_quadripartition_sets(**kwargs):
            # optional: sort qparts by len or min name (p3, p2), (p1, p0)
            # always: if not type==set sort within by name
            # returned as sets or sorted lists

            clade1 = format_quadripartition(below[0], below[1], type, sort)
            clade2 = format_quadripartition(other[0], other[1], type, sort)

            p1 = tformat(clade1[0])
            p2 = tformat(clade1[1])
            p3 = tformat(clade2[0])
            p4 = tformat(clade2[1])

            if sort:
                plen1 = len(p1) + len(p2)
                plen2 = len(p3) + len(p4)
                if plen1 > plen2:
                    p1, p2, p3, p4 = p3, p4, p1, p2 
                elif plen1 == plen2:
                    if min(min(p1), min(p2)) > min(min(p3), min(p4)):
                        p1, p2, p3, p4 = p3, p4, p1, p2
            if collapse:
                yield (p1, p2, p3, p4)
            else:
                yield ((p1, p2), (p3, p4))


def _build_node_names_for_sorting(node: Node) -> str:
    """Returns node name to use while sorting tip and internal nodes."""
    if node.is_leaf():
        return node.name
    return "".join(sorted(node.get_leaf_names())[::-1])


def format_quadripartition(
    below: Set,
    other: Set,
    type: Callable,
    sort: bool,
) -> Tuple[Sequence[Node], Sequence[Node]]:
    """Sort bipartitions ({Node, Node}, {Node, Node}).

    - First sorted by len: ({Node, Node}, {Node, Node, Node})
    - Then by min Node name: ({'a', 'z'}, {'b', 'x'})
    - which also requires assigning names to internal Nodes when present
    based on their descendant tips: ({'a', 'z', 'za'}, {'b', 'x', 'xb'})
    - if the type is not set then items within partitions are
    consistently sorted: (['a', 'z', 'za'], ['b', 'x', 'xb']) and can
    be converted to final type (e.g., tuple) in later steps.
    """
    blen = len(below)
    olen = len(other)

    # if type == set then items within a partition do not need to be
    # sorted, but part1 vs part2 does still need it.
    if sort and (type == set):
        if blen < olen:
            return below, other
        elif olen < blen:
            return other, below
        else:
            # min names
            bmin = min(i.name for i in below)
            omin = min(i.name for i in other)
            if bmin < omin:
                return below, other
            else:
                return other, below

    # if sort == False and type != set then items within partition need
    # to be sorted, but part1 vs part2 does not b/c we want to maintain
    # the (child, parent) order.
    if not sort:
        return (
            sorted(below, key=_build_node_names_for_sorting),
            sorted(other, key=_build_node_names_for_sorting),
        )

    # if sort == True and type != set then call sort on each partition
    # as well as between the two partitions.
    if blen < olen:
        return (
            sorted(below, key=_build_node_names_for_sorting),
            sorted(other, key=_build_node_names_for_sorting),
        )
    elif olen < blen:
        return (
            sorted(other, key=_build_node_names_for_sorting),
            sorted(below, key=_build_node_names_for_sorting),
        )
    else:
        # min names
        bmin = min(_build_node_names_for_sorting(i) for i in below)
        omin = min(_build_node_names_for_sorting(i) for i in other)
        if bmin < omin:
            return (
                sorted(below, key=_build_node_names_for_sorting),
                sorted(other, key=_build_node_names_for_sorting),
            )
        else:
            return (
                sorted(other, key=_build_node_names_for_sorting),
                sorted(below, key=_build_node_names_for_sorting),
            )


if __name__ == "__main__":

    import toytree

    # newick = "((a,b)X,((c,d)Y,e)Z)R;"
    # tree = toytree.tree(newick).root("c")

    # tree = toytree.rtree.baltree(12, seed=1234).unroot()
    tree = toytree.tree("(a,b,((c,d)CD,(e,f)EF)X)AB;")

    results = iter_quadripartitions(tree, type=list, sort=True, include_internal_nodes=True)
    
    for item in results:
        print(item[1][0][0])

    # tree._draw_browser(ts='r')
    # for bipart in _iter_quadripartition_sets(tree, include_internal_nodes=True):
    # for part in iter_quadripartitions(tree,):
    # print(part)

    # args to get consistently sorted biparts regardless of rooting
    # tree = tree.root("X")
    # for i in sorted(iter_quadripartitions(tree, type=set, sort=True)):
    #     print(i)
    # [(('a', 'b'), ('c', 'd', 'e', 'f')),
    #  (('c', 'd'), ('a', 'b', 'e', 'f')),
    #  (('e', 'f'), ('a', 'b', 'c', 'd'))]


    # tree = tree.root("X")
    # for i in sorted(iter_quadripartitions(tree, sort=True, collapse=True)):
    #     print(i)


    # convert consistent bipartitions to sets for easy comparison
    # x = set(tree.root('a').enum.iter_quadripartitions(type=tuple, sort=True))
    # y = set(tree.root('e').enum.iter_quadripartitions(type=tuple, sort=True))
    # assert x == y

    # tree = toytree.rtree.unittree(6, seed=123).unroot()
    # c1, *_ = tree.draw('p')

    # for bipart in _iter_quadripartition_sets(tree, 'idx'):#contract_partitions=True):
    #     print(bipart)

    # print("+========================================================")
    # tree = toytree.rtree.unittree(6, seed=123).root("r0")
    # c2, *_ = tree.draw(ts='p')

    # for bipart in _iter_quadripartition_sets(tree, 'idx'):#contract_partitions=True):
    #     print(bipart)

    # toytree.utils.show([c1, c2])
